# coding: utf-8

import numpy as np
from typing import List, Tuple
from .._global import OptionalModule
from .cameraConfigTools import Spot_boxes, Box

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")


class DISVE:
  """This tool computes the displacement of regions of interest (patches) on an
  image compared to a reference image.

  It relies on cross-correlation algorithms to calculate the displacement.
  Different algorithms are available depending on the needs.
  This tool is mainly used to perform video-extensometry on speckled surfaces,
  although it can as well be of use for other applications.
  """

  def __init__(self,
               patches: Spot_boxes,
               method: str = 'Disflow',
               alpha: float = 3,
               delta: float = 1,
               gamma: float = 0,
               finest_scale: int = 1,
               iterations: int = 1,
               gradient_iterations: int = 10,
               patch_size: int = 8,
               patch_stride: int = 3,
               border: float = 0.2,
               safe: bool = True,
               follow: bool = True) -> None:
    """Sets a few attributes and initializes Disflow if this method was
    selected.

    Args:
      patches: The regions to track, given as a :class:`Spot_boxes` object.
      method: The method to use to calculate the displacement. `Disflow` uses
        opencv's DISOpticalFlow and `Lucas Kanade` uses opencv's
        calcOpticalFlowPyrLK, while all other methods are based on a basic
        cross-correlation in the Fourier domain. `Pixel precision` calculates
        the displacement by getting the position of the maximum of the
        cross-correlation, and has thus a 1-pixel resolution. It is mainly
        meant for debugging. `Parabola` refines the result of
        `Pixel precision` by interpolating the neighborhood of the maximum, and
        have thus sub-pixel resolutions.
      alpha: Weight of the smoothness term in DisFlow.
      delta: Weight of the color constancy term in DisFlow.
      gamma: Weight of the gradient constancy term in DisFlow.
      finest_scale: Finest level of the Gaussian pyramid on which the flow
        is computed in DisFlow (`0` means full scale).
      iterations: Maximum number of gradient descent iterations in the
        patch inverse search stage in DisFlow.
      gradient_iterations: Maximum number of gradient descent iterations
        in the patch inverse search stage in DisFlow.
      patch_size: Size of an image patch for matching in DisFlow
        (in pixels).
      patch_stride: Stride between neighbor patches in DisFlow. Must be
        less than patch size.
      border: Crop the patch on each side according to this value before
        calculating the displacements. 0 means no cropping, 1 means the entire
        patch is cropped.
      safe: If :obj:`True`, checks whether the patches aren't exiting the
        image, and raises an error if that's the case.
      follow: It :obj:`True`, the patches will move to follow the displacement
        of the image.
    """

    # These attributes are accessed by the parent class
    self._patches = patches
    self._offsets = [(0, 0) for _ in patches]

    # Other attributes to set
    self._method = method
    self._border = border
    self._safe = safe
    self._follow = follow

    # These attributes will be set later
    self._img0 = None
    self._height, self._width = None, None

    # Initialize Disflow if it is the selected method
    if self._method == 'Disflow':
      self._dis = cv2.DISOpticalFlow_create(cv2.DISOPTICAL_FLOW_PRESET_FAST)
      self._dis.setVariationalRefinementIterations(iterations)
      self._dis.setVariationalRefinementAlpha(alpha)
      self._dis.setVariationalRefinementDelta(delta)
      self._dis.setVariationalRefinementGamma(gamma)
      self._dis.setFinestScale(finest_scale)
      self._dis.setGradientDescentIterations(gradient_iterations)
      self._dis.setPatchSize(patch_size)
      self._dis.setPatchStride(patch_stride)
    else:
      self._dis = None

  def set_img0(self, img0: np.ndarray) -> None:
    """Sets the reference image for the cross-correlation."""

    self._img0 = img0
    self._height, self._width, *_ = img0.shape

    # Now that there's an initial image, checking that the patches are valid
    if self._safe:
      self._check_offsets()

  def calculate_displacement(self, img: np.ndarray) -> List[float]:
    """Returns the displacement of every patch, calculated according to the
    chosen method.

    Also updates the patch offsets if required, and updates the window for
    following the patches if any.
    """

    # Making sure the reference image exists
    if self._img0 is None:
      raise ValueError("The method set_img0 must be called first for setting "
                       "the reference image !")

    # Compute the displacement for each patch
    displacements = []
    for patch, offset in zip(self._patches, self._offsets):

      if self._method == 'Disflow':
        displacements.append(self._calc_disflow(patch, img, offset))

      elif self._method == 'Pixel precision':
        displacements.append(self._calc_pixel_precision(patch, img, offset))

      elif self._method == 'Parabola':
        displacements.append(self._calc_parabola(patch, img, offset))

      elif self._method == 'Lucas Kanade':
        displacements.append(self._calc_lucas_kanade(patch, img, offset))

      else:
        raise ValueError("Wrong method specified !")

    # If required, updates the patch offsets
    if self._follow:
      for disp, (y_offset, x_offset), patch in zip(displacements,
                                                   self._offsets,
                                                   self._patches):

        patch.x_start = round(patch.x_start + disp[0])
        patch.x_end = round(patch.x_end + disp[0])
        patch.y_start = round(patch.y_start + disp[1])
        patch.y_end = round(patch.y_end + disp[1])

        disp[0] += x_offset
        disp[1] += y_offset

      self._offsets = [(round(y_disp), round(x_disp)) for
                       x_disp, y_disp in displacements]

      # Check that the patches are not exiting the image
      if self._safe:
        self._check_offsets()

    displacements = [coord for disp in displacements for coord in disp]
    return displacements

  def _calc_disflow(self,
                    patch: Box,
                    img: np.ndarray,
                    offset: Tuple[int, int]) -> List[float]:
    """Returns the displacement between the original and the current image with
    a sub-pixel precision, using Disflow."""

    disp_img = self._dis.calc(self._get_patch(self._img0, patch, offset),
                              self._get_patch(img, patch), None)
    return np.average(self._trim_patch(disp_img), axis=(0, 1)).tolist()

  def _calc_pixel_precision(self,
                            patch: Box,
                            img: np.ndarray,
                            offset: Tuple[int, int]) -> List[float]:
    """Returns the displacement between the original and the current image with
    a precision limited to 1 pixel."""

    cross_correl, max_width, max_height = self._cross_correlation(
      self._get_patch(self._img0, patch, offset), self._get_patch(img, patch))

    height, width = cross_correl.shape[0], cross_correl.shape[1]
    return [-(max_width - width / 2), -(max_height - height / 2)]

  def _calc_parabola(self,
                     patch: Box,
                     img: np.ndarray,
                     offset: Tuple[int, int]) -> List[float]:
    """Returns the displacement between the original and the current image with
    a sub-pixel precision, using two parabola fits (one in x and one in y)."""

    cross_correl, max_width, max_height = self._cross_correlation(
      self._get_patch(self._img0, patch, offset), self._get_patch(img, patch))

    height, width = cross_correl.shape[0], cross_correl.shape[1]
    y_disp = -(max_height - height / 2)
    x_disp = -(max_width - width / 2)

    x_disp -= self._parabola_fit(cross_correl[max_height,
                                 max_width - 1: max_width + 2])
    y_disp -= self._parabola_fit(cross_correl[max_height - 1: max_height + 2,
                                 max_width])

    return [x_disp, y_disp]

  def _calc_lucas_kanade(self,
                         patch: Box,
                         img: np.ndarray,
                         offset: Tuple[int, int]) -> List[float]:
    """Returns the displacement between the original and the current image with
    a sub-pixel precision, using the Lucas Kanade algorithm."""

    # Getting the center of the patch
    x_top, x_bottom, y_left, y_right = patch.sorted()
    center_y, center_x = (y_right - y_left) // 2, (x_bottom - x_top) // 2

    next_, _, _ = cv2.calcOpticalFlowPyrLK(
      self._get_patch(self._img0, patch, offset), self._get_patch(img, patch),
      np.array([[center_x, center_y]]).astype('float32'), None)
    new_x, new_y = np.squeeze(next_)

    return [new_x - center_x, new_y - center_y]

  @staticmethod
  def _parabola_fit(arr: np.ndarray) -> float:
    """Returns the abscissa of the maximum of a parabola defined by 3
    points in positions x=-1, x=0 and x=1.

    Args:
      arr: This array contains the y values for the 3 points.
    """

    return (arr[0] - arr[2]) / (2 * (arr[0] - 2 * arr[1] + arr[2]))

  @staticmethod
  def _cross_correlation(img0: np.ndarray,
                         img1: np.ndarray) -> (np.ndarray, int, int):
    """Performs a cross-correlation operation on two patches in the Fourier
    domain.

    Returns:
      The result of the correlation in the real domain as an image, as well as
      the position of the maximum of this image.
    """

    # Getting the size of the image
    height, width = img0.shape[0], img0.shape[1]

    # Find the closest power of 2 length from the image
    height_log2 = 2 ** int(np.log2(height + 0.5))
    width_log2 = 2 ** int(np.log2(width + 0.5))

    # Cropping the image to a power of 2
    img0 = img0[(height - height_log2) // 2: (height + height_log2) // 2,
                (width - width_log2) // 2: (width + width_log2) // 2]
    img1 = img1[(height - height_log2) // 2: (height + height_log2) // 2,
                (width - width_log2) // 2: (width + width_log2) // 2]

    # Convert to Fourier for fast cross-correlation
    img0_fourier = cv2.dft(img0.astype(np.float32),
                           flags=cv2.DFT_COMPLEX_OUTPUT)
    img1_fourier = cv2.dft(img1.astype(np.float32),
                           flags=cv2.DFT_COMPLEX_OUTPUT)

    # Compute cross-correlation by convolution
    cross_fourier = cv2.mulSpectrums(img0_fourier, img1_fourier,
                                     flags=0, conjB=True)

    # Convert back to physical space
    cross_shifted = cv2.idft(cross_fourier)
    # Un-shift after FFT
    cross_correl = np.fft.ifftshift(cross_shifted[:, :, 0])

    # Find the maximum of the cross-correlation
    max_width, max_height = cv2.minMaxLoc(cross_correl)[-1]
    # Keep the average in case several maxima found
    max_width, max_height = int(np.mean(max_width)), int(np.mean(max_height))

    return cross_correl, max_width, max_height

  @staticmethod
  def _get_patch(img: np.ndarray,
                 patch: Box,
                 offset: Tuple[int, int] = (0, 0)) -> np.ndarray:
    """Returns the part of the image corresponding to the given patch at the
    given offset."""

    y_off, x_off = offset
    x_top, x_bottom, y_left, y_right = patch.sorted()
    return np.array(img[y_left - y_off: y_right - y_off,
                        x_top - x_off: x_bottom - x_off])

  def _trim_patch(self, patch: np.ndarray) -> np.ndarray:
    """Trims the border of a patch according to the value set by the user, and
    returns the tuple corresponding to the trimmed patch."""

    height, width, *_ = patch.shape
    return patch[int(height * self._border / 2):
                 int(height * (1 - self._border / 2)),
                 int(width * self._border / 2):
                 int(width * (1 - self._border / 2))]

  def _check_offsets(self) -> None:
    """Check if the patches are still within the image, and raises an error if
    one of them is out."""

    for patch in self._patches:
      x_top, x_bottom, y_left, y_right = patch.sorted()

      # Checking the left border
      if x_top < 0:
        raise RuntimeError("Region exiting the ROI (left)")

      # Checking the right border
      elif x_bottom > self._width:
        raise RuntimeError("Region exiting the ROI (right)")

      # Checking the top border
      if y_left < 0:
        raise RuntimeError("Region exiting the ROI (top)")

      # Checking the bottom border
      elif y_right > self._height:
        raise RuntimeError("Region exiting the ROI (bottom)")



# coding: utf-8

from struct import pack, unpack
from typing import Union, Optional
import time
from .actuator import Actuator
from .._global import OptionalModule

try:
  import serial
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("pyserial")


def convert_to_byte(number: float, length: str) -> bytes:
  """This functions converts decimal into bytes.

  Mandatory in order to send or read anything into/from MAC Motors registers.
  """

  # get hex byte sequence in required '\xXX\xXX', big endian format.
  encoded = pack('%s' % length, number)
  # b = bytearray(encoded, 'hex')
  c = b''
  for i in range(0, len(encoded)):
    # x = encoded[0] ^ 0xff  # get the complement to 255
    # x = pack('B', x)  # byte formalism
    # concatenate byte and complement and add it to the sequence
    c += bytes([encoded[i], encoded[i] ^ 0xFF])
  return c


def convert_to_dec(sequence: bytes) -> float:
  """This functions converts bytes into decimals.

  Mandatory in order to send or read anything into/from MAC Motors registers.
  """

  # sequence=sequence[::2] ## cut off "complement byte"
  decim = unpack('i', sequence)  # convert to signed int value
  return decim[0]


class Biotens(Actuator):
  """Open the connection, and initialise the Biotens.

  Note:
    You should only use this class to communicate with the Biotens.
  """

  def __init__(self,
               port: str = '/dev/ttyUSB0',
               baudrate: int = 19200) -> None:
    """Sets the instance attributes.

    Args:
      port (:obj:`str`, optional): Path to connect to the serial port.
      baudrate (:obj:`int`, optional): Set the corresponding baud rate.
    """

    Actuator.__init__(self)
    self.port = port
    self.baudrate = baudrate

  def open(self) -> None:
    self.ser = serial.Serial(self.port, baudrate=19200, timeout=0.1)
    self.clear_errors()

  def reset_position(self) -> None:
    """Actuators goes out completely, in order to set the initial position."""

    init_position = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(38, 'B') +\
        convert_to_byte(4, 'B') +\
        convert_to_byte(0, 'i') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(38, 'B') + b'\xAA\xAA'

    init_speed = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(40, 'B') +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(-50, 'h') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(40, 'B') + b'\xAA\xAA'

    init_torque = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(41, 'B') +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(1023, 'i') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(41, 'B') +\
        b'\xAA\xAA'

    to_init = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(37, 'B') +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(0, 'h') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(37, 'B') + b'\xAA\xAA'

    self.ser.writelines([init_position, init_speed, init_torque, to_init])
    self.ser.write(b'\x52\x52\x52\xFF\x00' +
                   convert_to_byte(2, 'B') +
                   convert_to_byte(2, 'B') +
                   convert_to_byte(12, 'h') +
                   b'\xAA\xAA\x50\x50\x50\xFF\x00' +
                   convert_to_byte(2, 'B') +
                   b'\xAA\xAA')
    last_position_si = 0
    position_si = 99
    time.sleep(1)
    while position_si != last_position_si:
      last_position_si = position_si
      position_si = self.get_position()
      print("position : ", position_si)
    print("init done")
    self.stop()
    # time.sleep(1)
    # initializes the count when the motors is out.
    start_position = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(10, 'B') +\
        convert_to_byte(4, 'B') +\
        convert_to_byte(0, 'i') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(10, 'B') +\
        b'\xAA\xAA'
    self.ser.write(start_position)
    # time.sleep(1)
    try:
      self.ser.readlines()
    except serial.SerialException:
      pass

  def reset(self) -> None:
    """"""

    pass

  def stop(self) -> None:
    """Stop the motor."""

    command = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(0, 'h') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(2, 'B') + b'\xAA\xAA'
    self.ser.write(command)
    # return command

  def close(self) -> None:
    self.stop()
    self.ser.close()

  def clear_errors(self) -> None:
    """Clears error in motor registers."""

    command = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(35, 'B') +\
        convert_to_byte(4, 'B') +\
        convert_to_byte(0, 'i') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(35, 'B') + b'\xAA\xAA'
    self.ser.write(command)

  def set_speed(self, speed: float) -> None:
    """Pilot in speed mode, requires speed in `mm/min`."""

    # converts speed in motors value
    # displacement rate in mm/min, V_SOll in 1/16 encoder counts/sample.
    # 4096 encounter counts/revolution, sampling frequency = 520.8Hz,
    # screw thread=5.
    speed_soll = int(round(16 * 4096 * speed / (520.8 * 60 * 5)))
    set_speed = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(5, 'B') +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(speed_soll, 'h') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(5, 'B') + b'\xAA\xAA'

    # set torque to default value 1023
    set_torque = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(7, 'B') +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(1023, 'h') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(7, 'B') +\
        b'\xAA\xAA'

    # set acceleration to 10000 mm/s²
    # (default value, arbitrarily chosen, works great so far)
    asoll = int(round(16 * 4096 * 10000 / (520.8 * 520.8 * 5)))
    set_acceleration = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(6, 'B') +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(asoll, 'h') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(6, 'B') + b'\xAA\xAA+'

    command = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(1, 'h') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(2, 'B') +\
        b'\xAA\xAA'

    # write every parameters in motor's registers
    self.ser.writelines([set_speed, set_torque, set_acceleration, command])

  def set_position(self,
                   position: float,
                   speed: Optional[float] = None) -> None:
    """Pilot in position mode, needs speed and final position to run
    (in `mm/min` and `mm`)."""

    if speed is None:
      raise ValueError("The Biotens actuator needs both a position and a speed"
                       " command when driven in position mode !")

    # conversion of position from mm into encoder's count
    position_soll = int(round(position * 4096 / 5))
    set_position = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(3, 'B') +\
        convert_to_byte(4, 'B') +\
        convert_to_byte(position_soll, 'i') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(3, 'B') + b'\xAA\xAA+'

    # converts speed in motors value
    # displacement rate in mm/min, V_SOll in 1/16 encoder counts/sample.
    # 4096 encounter counts/revolution, sampling frequency = 520.8Hz
    # screw thread=5.
    speed_soll = int(round(16 * 4096 * speed / (520.8 * 60 * 5)))
    set_speed = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(5, 'B') +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(speed_soll, 'h') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(5, 'B') +\
        b'\xAA\xAA'

    # set torque to default value 1023
    set_torque = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(7, 'B') +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(1023, 'h') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(7, 'B') +\
        b'\xAA\xAA'

    # set acceleration to 10000 mm/s²
    # (default value, arbitrarily chosen, works great so far)
    asoll = int(round(16 * 4096 * 10000 / (520.8 * 520.8 * 5)))
    set_acceleration = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(6, 'B') +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(asoll, 'h') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(6, 'B') +\
        b'\xAA\xAA'

    command = b'\x52\x52\x52\xFF\x00' +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(2, 'B') +\
        convert_to_byte(2, 'h') +\
        b'\xAA\xAA\x50\x50\x50\xFF\x00' +\
        convert_to_byte(2, 'B') +\
        b'\xAA\xAA'

    # write every parameters in motor's registers
    self.ser.writelines([set_position, set_speed,
                        set_torque, set_acceleration, command])

  def get_position(self) -> float:
    """Reads current position.

    Returns:
      Current position of the motor.
    """

    for i in range(20):
      r = self._get_position()
      if r is not None:
        return r
      time.sleep(.01)
    raise IOError("Could not read biotens pos!")

  def _get_position(self) -> Union[float, None]:
    try:
      self.ser.readlines()
    except serial.SerialException:
      # print "readlines failed"
      pass
    # print "position read"
    command = b'\x50\x50\x50\xFF\x00' + convert_to_byte(10, 'B') + b'\xAA\xAA'

    self.ser.write(command)
    # time.sleep(0.01)
    # print "reading..."
    # print self.ser.inWaiting()
    position_ = self.ser.read(19)
    if len(position_) != 19:
      return None
    # print "read"
    position = position_[9:len(position_) - 2:2]
    position = convert_to_dec(position) * 5 / 4096.
    return position



# coding: utf-8

from multiprocessing import Process, Pipe
from multiprocessing.connection import Connection
from typing import Optional, Tuple, List
import numpy as np
from itertools import combinations
from time import sleep
from .._global import OptionalModule
from .cameraConfigBoxes import Spot_boxes, Box

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")
try:
  from skimage.filters import threshold_otsu
  from skimage.morphology import label
  from skimage.measure import regionprops
except (ModuleNotFoundError, ImportError):
  label = OptionalModule("skimage", "Please install scikit-image to use"
                         "Video-extenso")
  threshold_otsu = regionprops = label


class LostSpotError(Exception):
  """Exception raised when a spot is lost, or when there's too much
  overlapping."""

  pass


class VideoExtenso:
  """This class tracks up to 4 spots on successive images, and computes the
  strain on the material based on the displacement of the spots from their
  original position.

  The first step is to detect the spots to track. Once this is done, the
  initial distances in x and y between the spots is saved. For each spot, a
  Process in charge of tracking it is started. When a new image is received,
  subframes based on the last known positions of the spots are sent to the
  tracker processes, and they return the new positions of the detected spots.
  Based on these new positions, updated strain values are returned.

  It is possible to track only one spot, in which case only the position of its
  center is returned and the strain values are left to 0.
  """

  def __init__(self,
               white_spots: bool = False,
               update_thresh: bool = False,
               num_spots: Optional[int] = None,
               safe_mode: bool = False,
               border: int = 5,
               min_area: int = 150,
               blur: Optional[int] = 5) -> None:
    """Sets the args and the other instance attributes.

    Args:
      white_spots: If :obj:`True`, detects white objects on a black background,
        else black objects on a white background.
      update_thresh: If :obj:`True`, the threshold for detecting the spots is
        re-calculated for each new image. Otherwise, the first calculated
        threshold is kept for the entire test. The spots are less likely to be
        lost with adaptive threshold, but the measurement will be more noisy.
        Adaptive threshold may also yield inconsistent results when spots are
        lost.
      num_spots: The number of spots to detect, between 1 and 4. The class will
        then try to detect this exact number of spots, and won't work if not
        enough spots can be found. If this argument is not given, at most 4
        spots can be detected but the class will work with any number of
        detected spots between 1 and 4.
      safe_mode: If :obj:`True`, the class will stop and raise an exception as
        soon as overlapping is detected. Otherwise, it will first try to reduce
        the detection window to get rid of overlapping. This argument should be
        used when inconsistency in the results may have critical consequences.
      border: When searching for the new position of a spot, the class will
        search in the last known bounding box of this spot plus a few
        additional pixels in each direction. This argument sets the number of
        additional pixels to use. It should be greater than the expected
        "speed" of the spots, in pixels / frame. But if it's too big, noise or
        other spots might hinder the detection.
      min_area: The minimum area an object should have to be potentially
        detected as a spot. The value is given in pixels, as a surface unit.
        It must of course be adapted depending on the resolution of camera and
        the size of the spots to detect.
      blur: The size in pixels of the kernel to use for applying a median blur
        to the image before the spot detection. If not given, no blurring is
        performed. A slight blur improves the spot detection by smoothening the
        noise, but also takes a bit more time compared to no blurring.
    """

    if num_spots is not None and num_spots not in range(1, 5):
      raise ValueError("num_spots should be either None, 1, 2, 3 or 4 !")
    self._num_spots = num_spots

    # Setting the args
    self._white_spots = white_spots
    self._update_thresh = update_thresh
    self._safe_mode = safe_mode
    self._border = border
    self._min_area = min_area
    self._blur = blur

    # Setting the other attributes
    self._consecutive_overlaps = 0
    self.spots = Spot_boxes()
    self._thresh = 0 if white_spots else 255
    self.x_l0 = None
    self.y_l0 = None
    self._trackers = list()
    self._pipes = list()

  def __del__(self) -> None:
    """Security to ensure there are no zombie processes left when exiting."""

    self.stop_tracking()

  def detect_spots(self,
                   img: np.ndarray,
                   y_orig: int,
                   x_orig: int) -> Optional[Spot_boxes]:
    """Transforms the image to improve spot detection, detects up to 4 spots
    and return a Spot_boxes object containing all the detected spots.

    Args:
      img: The sub-image on which the spots should be detected.
      y_orig: The y coordinate of the top-left pixel of the sub-image on the
        entire image.
      x_orig: The x coordinate of the top-left pixel of the sub-image on the
        entire image.

    Returns:
      A Spot_boxes object containing all the detected spots.
    """

    # First, blurring the image if asked to
    if self._blur is not None and self._blur > 1:
      img = cv2.medianBlur(img, self._blur)

    # Determining the best threshold for the image
    self._thresh = threshold_otsu(img)

    # Getting all pixels superior or inferior to threshold
    if self._white_spots:
      black_white = img > self._thresh
    else:
      black_white = img <= self._thresh

    # The original image is not needed anymore
    del img

    # Detecting the spots on the image
    props = regionprops(label(black_white))

    # The black and white image is not needed anymore
    del black_white

    # Removing regions that are too small or not circular
    props = [prop for prop in props if prop.solidity > 0.8
             and prop.area > self._min_area]

    # Detecting overlapping spots and storing the ones that should be removed
    to_remove = list()
    for prop_1, prop_2 in combinations(props, 2):
      if self._overlap_bbox(prop_1, prop_2):
        to_remove.append(prop_2 if prop_1.area > prop_2.area else prop_1)

    # Now removing the overlapping spots
    if to_remove:
      print("[VideoExtenso] Overlapping spots found, removing the smaller "
            "overlapping ones")
    for prop in to_remove:
      props.remove(prop)

    # Sorting the regions by area
    props = sorted(props, key=lambda prop: prop.area, reverse=True)

    # Keeping only the biggest areas to match the target number of spots
    if self._num_spots is not None:
      props = props[:self._num_spots]
    else:
      props = props[:4]

    # Indicating the user if not enough spots were found
    if not props:
      print("[VideoExtenso] No spots found !")
      return
    elif self._num_spots is not None and len(props) != self._num_spots:
      print(f"[VideoExtenso] Expected {self._num_spots} spots, "
            f"found only {len(props)}")
      return

    # Replacing the previously detected spots with the new ones
    self.spots.reset()
    for i, prop in enumerate(props):
      # Extracting the properties of interest
      y, x = prop.centroid
      y_min, x_min, y_max, x_max = prop.bbox
      # Adjusting with the offset given as argument
      x += x_orig
      x_min += x_orig
      x_max += x_orig
      y += y_orig
      y_min += y_orig
      y_max += y_orig

      self.spots[i] = Box(x_start=x_min, x_end=x_max,
                          y_start=y_min, y_end=y_max,
                          x_centroid=x, y_centroid=y)

    return self.spots

  def save_length(self) -> bool:
    """Saves the initial length in x and y between the detected spots."""

    # Cannot determine a length if no spots was detected
    if self.spots.empty():
      print("[VideoExtenso] Cannot save L0, no spots selected yet !")
      return False

    # Simply taking the distance between the extrema as the initial length
    if len(self.spots) > 1:
      x_centers = [spot.x_centroid for spot in self.spots if spot is not None]
      y_centers = [spot.y_centroid for spot in self.spots if spot is not None]
      self.x_l0 = max(x_centers) - min(x_centers)
      self.y_l0 = max(y_centers) - min(y_centers)

    # If only one spot detected, setting the initial lengths to 0
    else:
      self.x_l0 = 0
      self.y_l0 = 0

    return True

  def start_tracking(self) -> None:
    """Creates a Tracker process for each detected spot, and starts it.

    Also creates a Pipe for each spot to communicate with the Tracker process.
    """

    if self.spots.empty():
      raise AttributeError("[VideoExtenso] No spots selected, aborting !")

    for spot in self.spots:
      if spot is None:
        continue

      inlet, outlet = Pipe()
      tracker = Tracker(pipe=outlet,
                        white_spots=self._white_spots,
                        thresh=None if self._update_thresh else self._thresh,
                        blur=self._blur)
      self._pipes.append(inlet)
      self._trackers.append(tracker)
      tracker.start()

  def stop_tracking(self) -> None:
    """Stops all the active Tracker processes, either gently or by terminating
    them if they don't stop by themselves."""

    if any((tracker.is_alive() for tracker in self._trackers)):
      # First, gently asking the trackers to stop
      for pipe, tracker in zip(self._pipes, self._trackers):
        if tracker.is_alive():
          pipe.send(('stop', 'stop', 'stop'))
      sleep(0.05)

      # If they're not stopping, killing the trackers
      for tracker in self._trackers:
        if tracker.is_alive():
          tracker.terminate()

  def get_data(self,
               img: np.ndarray) -> Optional[Tuple[List[Tuple[float, ...]],
                                                  float, float]]:
    """Takes an image as an input, performs spot detection on it, computes the
    strain from the newly detected spots, and returns the spot positions and
    strain values.

    Args:
      img: The image on which the spots should be detected.

    Returns:
      A :obj:`list` containing tuples with the coordinates of the centers of
      the detected spots, and the calculated x and y strain values.
    """

    # First, saving the initial length if not already saved
    if self.x_l0 is None or self.y_l0 is None:
       if not self.save_length():
         # If no spots are selected, it should be detected here
         raise IOError("[VideoExtenso] No spots selected, aborting !")

    # Sending the latest sub-image containing the spot to track
    # Also sending the coordinates of the top left pixel
    for pipe, spot in zip(self._pipes, self.spots):
      x_top, x_bottom, y_left, y_right = spot.sorted()
      slice_y = slice(max(0, y_left - self._border),
                      min(img.shape[0], y_right + self._border))
      slice_x = slice(max(0, x_top - self._border),
                      min(img.shape[1], x_bottom + self._border))
      pipe.send((slice_y.start, slice_x.start, img[slice_y, slice_x]))

    for i, (pipe, spot) in enumerate(zip(self._pipes, self.spots)):

      # Receiving the data from the tracker, if there's any
      if pipe.poll(timeout=0.5):
        box = pipe.recv()

        # In case a tracker faced an error, stopping them all and raising
        if isinstance(box, str):
          self.stop_tracking()
          raise LostSpotError(f"[VideoExtenso] Tracker returned error : {box}")

        self.spots[i] = box

    overlap = False

    # Checking if the newly received boxes overlaps with each other
    for box_1, box_2 in combinations(self.spots, 2):
      if self._overlap_box(box_1, box_2):

        # If there's overlapping in safe mode, raising directly
        if self._safe_mode:
          self.stop_tracking()
          raise LostSpotError("[VideoExtenso] Overlapping detected in safe"
                              " mode, raising exception directly")

        print("[VideoExtenso] Overlapping detected ! Reducing spot window")

        # If we're not in safe mode, simply reduce the boxes by 1 pixel
        # Also, make sure the box is not being reduced too much
        overlap = True
        x_top_1, x_bottom_1, y_left_1, y_right_1 = box_1.sorted()
        x_top_2, x_bottom_2, y_left_2, y_right_2 = box_2.sorted()
        box_1.x_start = min(x_top_1 + 1, box_1.x_centroid - 2)
        box_1.y_start = min(y_left_1 + 1, box_1.y_centroid - 2)
        box_1.x_end = max(x_bottom_1 - 1, box_1.x_centroid + 2)
        box_1.y_end = max(y_right_1 - 1, box_1.y_centroid + 2)
        box_2.x_start = min(x_top_2 + 1, box_2.x_centroid - 2)
        box_2.y_start = min(y_left_2 + 1, box_2.y_centroid - 2)
        box_2.x_end = max(x_bottom_2 - 1, box_2.x_centroid + 2)
        box_2.y_end = max(y_right_2 - 1, box_2.y_centroid + 2)

    if overlap:
      self._consecutive_overlaps += 1
      if self._consecutive_overlaps > 10:
        raise LostSpotError("[VideoExtenso] Too many consecutive "
                            "overlaps !")
    else:
      self._consecutive_overlaps = 0

    # If there are multiple spots, the x and y strains can be computed
    if len(self.spots) > 1:
      x = [spot.x_centroid for spot in self.spots if spot is not None]
      y = [spot.y_centroid for spot in self.spots if spot is not None]
      try:
        # The strain is calculated based on the positions of the extreme
        # spots in each direction
        exx = ((max(x) - min(x)) / self.x_l0 - 1) * 100
        eyy = ((max(y) - min(y)) / self.y_l0 - 1) * 100
      except ZeroDivisionError:
        # Shouldn't happen but adding a safety just in case
        exx, eyy = 0, 0
      centers = list(zip(y, x))
      return centers, eyy, exx

    # If only one spot was detected, the strain isn't computed
    else:
      x = self.spots[0].x_centroid
      y = self.spots[0].y_centroid
      return [(y, x)], 0, 0

  @staticmethod
  def _overlap_box(box_1: Box, box_2: Box) -> bool:
    """Determines whether two boxes are overlapping or not."""

    x_min_1, x_max_1, y_min_1, y_max_1 = box_1.sorted()
    x_min_2, x_max_2, y_min_2, y_max_2 = box_2.sorted()

    return max((min(x_max_1, x_max_2) - max(x_min_1, x_min_2)), 0) * max(
      (min(y_max_1, y_max_2) - max(y_min_1, y_min_2)), 0) > 0

  @staticmethod
  def _overlap_bbox(prop_1, prop_2) -> bool:
    """Determines whether two bboxes are overlapping or not."""

    y_min_1, x_min_1, y_max_1, x_max_1 = prop_1.bbox
    y_min_2, x_min_2, y_max_2, x_max_2 = prop_2.bbox

    return max((min(x_max_1, x_max_2) - max(x_min_1, x_min_2)), 0) * max(
      (min(y_max_1, y_max_2) - max(y_min_1, y_min_2)), 0) > 0


class Tracker(Process):
  """Process whose task is to track a spot on an image.

  It receives a subframe centered on the last known position of the spot, and
  returns the updated position of the detected spot.
  """

  def __init__(self,
               pipe: Connection,
               white_spots: bool = False,
               thresh: Optional[int] = None,
               blur: Optional[float] = 5) -> None:
    """Sets the args.

    Args:
      pipe: The :obj:`multiprocessing.connection.Connection` object through
        which the image is received and the updated coordinates of the spot
        are sent back.
      white_spots: If :obj:`True`, detects white objects on a black background,
        else black objects on a white background.
      thresh: If given, this threshold value will always be used for isolating
        the spot from the background. If not given (:obj:`None`), a new
        threshold is recalculated for each new subframe. Spots are less likely
        to be lost with an adaptive threshold, but it takes a bit more time.
      blur: If not :obj:`None`, the subframe is first blurred before trying to
        detect the spot. This argument gives the size of the kernel to use for
        blurring. Better results are obtained with blurring, but it takes a bit
        more time.
    """

    super().__init__()

    self._pipe = pipe
    self._white_spots = white_spots
    self._thresh = thresh
    self._blur = blur

    self._n = 0

  def run(self) -> None:
    """Continuously reads incoming subframes, tries to detect a spot and sends
    back the coordinates of the detected spot.

    Can only be stopped either with a :exc:`KeyboardInterrupt` or when
    receiving a text message from the parent VideoExtenso class.
    """

    # Looping forever for receiving data
    try:
      while True:
        # Making sure the call to recv is not blocking
        if self._pipe.poll(0.5):
          y_start, x_start, img = self._pipe.recv()
          self._n += 1

          # If a string is received, always means the process has to stop
          if isinstance(img, str):
            break

          # Simply sending back the new Box containing the spot
          try:
            self._pipe.send(self._evaluate(x_start, y_start, img))

          # If the caught exception is a KeyboardInterrupt, simply stopping
          except KeyboardInterrupt:
            break
          # Sending back the exception if anything else unexpected happened
          except (Exception,) as exc:
            self._pipe.send(str(exc))

    # In case the user presses CTRL+C, simply stopping the process
    except KeyboardInterrupt:
      pass

  def _evaluate(self, x_start: int, y_start: int, img: np.ndarray) -> Box:
    """Takes a sub-image, applies a threshold on it and tries to detect the new
    position of the spot.

    Args:
      x_start: The x position of the top left pixel of the subframe on the
        entire image.
      y_start: The y position of the top left pixel of the subframe on the
        entire image.
      img: The subframe on which to search for a spot.

    Returns:
      A Box object containing the x and y start and end positions of the
      detected spot, as well as the coordinates of the centroid.
    """

    # First, blurring the image if asked to
    if self._blur is not None and self._blur > 1:
      img = cv2.medianBlur(img, self._blur)

    # Determining the best threshold for the image if required
    thresh = self._thresh if self._thresh is not None else threshold_otsu(img)

    # Getting all pixels superior or inferior to threshold
    if self._white_spots:
      black_white = (img > thresh).astype('uint8')
    else:
      black_white = (img <= thresh).astype('uint8')

    # Checking that the detected spot is large enough
    if np.count_nonzero(black_white) < 0.1 * img.size:

      # If the threshold is pre-defined, trying again with an updated one
      if self._thresh is not None:
        print("[VideoExtenso Tracker] Detected spot too small compared with "
              "overall box size, recalculating threshold")
        thresh = threshold_otsu(img)
        if self._white_spots:
          black_white = (img > thresh).astype('uint8')
        else:
          black_white = (img <= thresh).astype('uint8')

        # If the spot still cannot be detected, aborting
        if np.count_nonzero(black_white) < 0.1 * img.size:
          raise LostSpotError("Couldn't detect spot with adaptive threshold, "
                              "aborting !")

      # If an adaptive threshold is already used, nothing more can be done
      else:
        raise LostSpotError("Couldn't detect spot with adaptive threshold, "
                            "aborting !")

    # Calculating the coordinates of the centroid using the image moments
    moments = cv2.moments(black_white)
    try:
      x = moments['m10'] / moments['m00']
      y = moments['m01'] / moments['m00']
    except ZeroDivisionError:
      raise ZeroDivisionError("Couldn't compute the centroid because the "
                              "moment of order 0, 0 is zero !")

    # Getting the updated centroid and coordinates of the spot
    x_min, y_min, width, height = cv2.boundingRect(black_white)
    return Box(x_start=x_start + x_min,
               y_start=y_start + y_min,
               x_end=x_start + x_min + width,
               y_end=y_start + y_min + height,
               x_centroid=x_start + x,
               y_centroid=y_start + y)



# coding: utf-8

from .block import Block
from .._global import OptionalModule
from typing import Dict, List, Union, Tuple, Any, Optional
from time import time, sleep
from subprocess import Popen, PIPE, STDOUT, TimeoutExpired
from threading import Thread
from queue import Queue, Empty
from ast import literal_eval
from pickle import loads, dumps, UnpicklingError
from socket import timeout, gaierror
from itertools import chain

try:
  import paho.mqtt.client as mqtt
except (ModuleNotFoundError, ImportError):
  mqtt = OptionalModule("paho.mqtt.client")

topics_type = List[Union[str, Tuple[str, ...]]]


class Client_server(Block):
  """Block for exchanging data on a local network using the MQTT protocol.

  This block can send data to an MQTT broker, receive data from this broker by
  subscribing to its topics, and also launch the Mosquitto broker.
  """

  def __init__(self,
               broker: bool = False,
               address: Any = 'localhost',
               port: int = 1883,
               init_output: Optional[Dict[str, Any]] = None,
               topics: Optional[topics_type] = None,
               cmd_labels: Optional[topics_type] = None,
               labels_to_send: Optional[topics_type] = None,
               verbose: bool = False,
               freq: float = 200,
               spam: bool = False) -> None:
    """Checks arguments validity and sets the instance attributes.

    Args:
      broker: If :obj:`True`, starts the Mosquitto broker during the prepare
        loop and stops it during the finish loop. If Mosquitto is not installed
        a :exc:`FileNotFoundError` is raised.
      address (optional): The network address on which the MQTT broker is
        running.
      port (:obj:`int`, optional): A network port on which the MQTT broker is
        listening.
      init_output (:obj:`dict`, optional): A :obj:`dict` containing for labels
        in ``topics`` the first value to be sent in the output links. Should be
        given in case the data comes from several sources and data for all
        labels may not be available during the first loops.
      topics (:obj:`list`, optional): A :obj:`list` of :obj:`str` and/or 
        :obj:`tuple` of :obj:`str`. Each string corresponds to the name of a 
        crappy label to be received from the broker. Each element of the list 
        is considered to be the name of an MQTT topic, to which the client 
        subscribes. After a message has been received on that topic, the block 
        returns for each label in the topic (i.e. each string in the tuple) the 
        corresponding data from the message. It also returns the current
        timestamp in the label `'t(s)'`.
      cmd_labels (:obj:`list`, optional): A :obj:`list` of :obj:`str` and/or 
        :obj:`tuple` of :obj:`str`. Each string corresponds to the name of a 
        crappy label to send to the broker. Each element of the list is 
        considered to be the name of an MQTT topic, in which the client 
        publishes. Grouping labels in a same topic (i.e. strings in a same 
        tuple) allows to keep the synchronization between signals coming from a 
        same block, as they will be published together in a same message. This 
        is mostly useful for sending a signal along with its timeframe.
      labels_to_send (:obj:`list`, optional): A :obj:`list` of :obj:`str` 
        and/or :obj:`tuple` of :obj:`str`. Allows to rename the labels before 
        publishing data. The structure of ``labels_to_send`` should be the 
        exact same as ``cmd_labels``, with each label in ``labels_to_send`` 
        replacing the corresponding one in ``cmd_labels``. This is especially 
        useful for transferring several signals along with their timestamps, as 
        the label ``'t(s)'`` should not appear more than once in the topics 
        list of the receiving block.
      verbose: If :obj:`True`, displays the looping frequency of the block.
      freq: The block will try to loop at this frequency.
      spam: If :obj:`True`, sends the last received values at each loop even if
        no new values were received from the broker.

    Note:
      - ``broker``:
        In order for the block to run, an MQTT broker must be running at the
        specified address on the specified port. If not, an
        :exc:`ConnectionRefusedError` is raised. The broker can be started and
        stopped manually by the user independently of the execution of crappy.
        It also doesn't need to be Mosquitto, any other MQTT broker can be
        used.

      - ``topics``:
        The presence of the same label in multiple topics will most likely lead 
        to a data loss.

      - ``cmd_labels``:
        It is not possible to group signals coming from different blocks in a
        same topic.

      - ``labels_to_send``:
        Differences in the structure of ``labels_to_send`` and ``cmd_labels``
        will not always raise an error, but may lead to a data loss.

      - **Single-value tuples**:
        Single-value tuples can be shortened as strings.
        ::

          topics=[('cmd1',), ('cmd2',)]
          cmd_labels=[('cmd1',), ('cmd2',)]
          labels_to_send=[('cmd1',), ('cmd2',)]

        is equivalent to
        ::

          topics=['cmd1', 'cmd2']
          cmd_labels=['cmd1', 'cmd2']
          labels_to_send=['cmd1', 'cmd2']


    Examples:
      - ``topics``:
        If
        ::

          topics=[('t1', 'cmd1'), 'sign']

        the client will subscribe to the topics
        ::

          ('t1', 'cmd1')
          ('sign',)

        The block will return data associated with the labels
        ::

          't1', 'cmd1'
          'sign'


      - ``cmd_labels``:
        If
        ::

          cmd_labels=[('t1', 'cmd1'), 'sign']

        the client will publish data in the form of
        ::

          [[t1_0, cmd1_0], [t1_1, cmd1_1], ...]
          [[sign_0], [sign_1], ...]

        in the topics
        ::

           ('t1', 'cmd1')
           ('sign',)


      - ``labels_to_send``:
        If
        ::

          cmd_labels=[('t(s)', 'cmd'), 'sign']
          labels_to_send=[('t1', 'cmd1'), 'sign']

        the data from labels
        ::

          't(s)', 'cmd'

        will be published in the topic
        ::

          ('t1', 'cmd1')

        and the data from label
        ::

          'sign'

        in the topic
        ::

          ('sign',)

    """

    Block.__init__(self)
    self.niceness = -10
    self.verbose = verbose
    self.freq = freq

    # Setting the args
    self._broker = broker
    self._address = address
    self._port = port
    self._spam = spam
    self._init_output = init_output if init_output is not None else dict()
    self._reader = Thread(target=self._output_reader)

    self._stop_mosquitto = False

    # Instantiating the client
    self._client = mqtt.Client(str(time()))
    self._client.on_connect = self._on_connect
    self._client.on_message = self._on_message
    self._client.reconnect_delay_set(max_delay=10)

    if topics is None and cmd_labels is None:
      print("[Client_server] WARNING: client-server block is neither an input "
            "nor an output !")

    # Preparing for receiving data
    if topics is not None:
      # Replacing strings with tuples
      self._topics = [topic if isinstance(topic, tuple) else (topic,) for
                      topic in topics]

      # The last out vals are given for each label, not each topic
      self._last_out_val = {label: None for label in chain(*self._topics)}
      print(self._last_out_val)

      # The buffer for received data is a dictionary of queues
      self._buffer_output = {topic: Queue() for topic in topics}

    else:
      self._topics = None
      self._buffer_output = None

    # Preparing for publishing data
    if cmd_labels is not None:
      # Replacing strings with tuples
      self._cmd_labels = [topic if isinstance(topic, tuple) else (topic,)
                          for topic in cmd_labels]

      if labels_to_send is not None:
        # Replacing strings with tuples
        labels_to_send = [topic if isinstance(topic, tuple) else (topic,)
                          for topic in labels_to_send]

        # Making sure the labels to send have the correct syntax
        if len(labels_to_send) != len(cmd_labels):
          raise ValueError("Either a label_to_send should be given for "
                           "every cmd_label, or none should be given ")

        # Preparing to rename labels to send using a dictionary
        self._labels_to_send = {cmd_label: label_to_send for
                                cmd_label, label_to_send in
                                zip(self._cmd_labels, labels_to_send)}
    else:
      self._cmd_labels = None
      self._labels_to_send = None

  def prepare(self) -> None:
    """Starts the broker and connects to it."""

    # Making sure the necessary inputs and outputs are present
    if self._topics is not None and not self.outputs:
      raise ValueError("topics are specified but there's no output link !")
    if self._cmd_labels is not None and not self.inputs:
      raise ValueError("cmd_labels are specified but there's no input link !")

    # Starting the broker
    if self._broker:
      self._launch_mosquitto()
      self._reader.start()
      sleep(2)
      print('[Client_server] Waiting for Mosquitto to start')
      sleep(2)

    # Connecting to the broker
    try_count = 15
    while True:
      try:
        self._client.connect(self._address, port=self._port, keepalive=10)
        break
      except timeout:
        print("[Client_server] Impossible to reach the given address, "
              "aborting")
        raise
      except gaierror:
        print("[Client_server] Invalid address given, please check the "
              "spelling")
        raise
      except ConnectionRefusedError:
        try_count -= 1
        if try_count == 0:
          print("[Client_server] Connection refused, the broker may not be "
                "running or you may not have the rights to connect")
          raise
        sleep(1)

    self._client.loop_start()

  def loop(self) -> None:
    """Receives data from the broker and/or sends data to the broker.

    The received data is then sent to the crappy blocks connected to this one.
    """

    """Loop for receiving data
    Each queue in the buffer is checked once: if not empty then the first list 
    of data is popped. This data is then associated to the corresponding 
    labels in dict_out. dict_out is finally returned if not empty. All the 
    labels should be returned at each loop iteration, so a buffer stores the 
    last value for each label and returns it if no other value was received. In 
    case no value was received yet for a given label, the user can also provide 
    init values to be sent at the beginning."""
    if self._topics is not None:
      dict_out = {}
      for topic in self._buffer_output:
        if not self._buffer_output[topic].empty():
          try:
            data_list = self._buffer_output[topic].get_nowait()
            for label, data in zip(topic, data_list):
              dict_out[label] = data
          except Empty:
            pass
      # Updating the _last_out_val buffer, and completing dict_out before
      # sending data if necessary
      if dict_out or (self._spam and all(val is not None for val in
                                         self._last_out_val.values())):
        for topic in self._buffer_output:
          for label in topic:
            if label not in dict_out:
              if self._last_out_val[label] is not None:
                dict_out[label] = self._last_out_val[label]
              elif label in self._init_output:
                dict_out[label] = self._init_output[label]
              else:
                raise ValueError(f"No value received for the topic {label} and"
                                 f" no init value given !")
            else:
              self._last_out_val[label] = dict_out[label]
        # Adding the timestamp before sending
        dict_out['t(s)'] = time() - self.t0
        self.send(dict_out)

    """Loop for sending data
    Data is first received as a list of dictionaries. For each topic, trying to 
    find a dictionary containing all the corresponding labels. Once this 
    dictionary has been found, its data is published as a list of list of 
    values."""
    if self._cmd_labels is not None:
      received_data = [link.recv_chunk(blocking=False) for link in self.inputs]
      for topic in self._cmd_labels:
        for dic in received_data:
          if dic is not None and all(label in dic for label in topic):
            self._client.publish(
              topic=str(self._labels_to_send[topic]) if
              self._labels_to_send is not None else str(topic),
              payload=dumps([dic[label] for label in topic]),
              qos=0)
            break

  def finish(self) -> None:
    """Disconnects from the broker and stops it."""

    # Disconnecting from the broker
    self._client.loop_stop()
    self._client.disconnect()

    # Stopping the broker
    if self._broker:
      try:
        self._proc.terminate()
        self._proc.wait(timeout=15)
        print('[Client_server] Mosquitto terminated with return code',
              self._proc.returncode)
        self._stop_mosquitto = True
        self._reader.join()
      except TimeoutExpired:
        print('[Client_server] Subprocess did not terminate in time')
        self._proc.kill()

  def _launch_mosquitto(self) -> None:
    """Starts Mosquitto in a subprocess."""

    try:
      self._proc = Popen(['mosquitto', '-p', str(self._port)],
                         stdout=PIPE,
                         stderr=STDOUT)
    except FileNotFoundError:
      print("[Client_server] Mosquitto is not installed !")
      raise

  def _output_reader(self) -> None:
    """Reads the output strings from Mosquitto's subprocess."""

    while not self._stop_mosquitto:
      for line in iter(self._proc.stdout.readline, b''):
        print('[Mosquitto] {0}'.format(line.decode('utf-8')), end='')
        if 'Error: Address already in use' in line.decode('utf-8'):
          print('Mosquitto is already running on this port')
      sleep(0.1)

  def _on_message(self, _, __, message) -> None:
    """Buffers the received data.

    The received message consists in a list of lists of values. Data is placed
    in the right buffer according to the topic, in the form of lists of values.
    """

    try:
      for data_points in zip(*loads(message.payload)):
        self._buffer_output[literal_eval(message.topic)].put_nowait(
          list(data_points))
    except UnpicklingError:
      print("[Client_server] Warning ! Message raised UnpicklingError, "
            "ignoring it")

  def _on_connect(self, _, __, ___, rc: Any) -> None:
    """Automatically subscribes to the topics when connecting to the broker."""

    print("[Client_server] Connected with result code " + str(rc))

    # Subscribing on connect, so that it automatically resubscribes when
    # reconnecting after a connection loss
    if self._topics is not None:
      for topic in self._topics:
        self._client.subscribe(topic=str(topic), qos=0)
        print("[Client_server] Subscribed to", topic)

    self._client.loop_start()



# coding: utf-8

"""
Demonstrates the use of the Mean block.

Not to be mistaken with ``modifier.Mean`` ! The Mean blocks performs the mean
at a given frequency, whereas the modifier will perform the average over a
chosen number of points regardless of the frequency.

No required hardware.
"""

import crappy

if __name__ == "__main__":
  g1 = crappy.blocks.Generator(
      [dict(type='sine', freq=2, amplitude=2, condition=None)],
      freq=200,
      cmd_label='cmd1'
    )

  g2 = crappy.blocks.Generator(
      [dict(type='sine', freq=.2, amplitude=2, condition=None)],
      freq=200,
      cmd_label='cmd2'
    )

  # Return a point every .5s
  # It will be the average received value
  m = crappy.blocks.Mean_block(.5)  # , out_labels=['cmd1', 'cmd2'])

  crappy.link(g1, m)
  crappy.link(g2, m)

  g = crappy.blocks.Grapher(('t(s)', 'cmd1'), ('t(s)', 'cmd2'))
  crappy.link(m, g)
  crappy.start()



# coding: utf-8

import numpy as np
from typing import Callable, Optional, List, Union, Tuple, Dict, Any
from pathlib import Path

from .._global import OptionalModule
from ..tool import GPUCorrel as GPUCorrel_tool
from .camera import Camera

try:
  import pycuda.tools
  import pycuda.driver
except (ModuleNotFoundError, ImportError):
  pycuda = OptionalModule("pycuda")


class GPUVE(Camera):
  """This block tracks patches on an image using GPU-accelerated Dense Inverse
  Search.

  The patches must be given explicitly as arguments, they cannot be selected in
  a GUI. The block sends the updated positions of the tracked patches to the
  downstream blocks, as well as the timestamp. It is roughly equivalent to the
  :ref:`Disve` block, but with GPU acceleration.
  """

  def __init__(self,
               camera: str,
               patches: List[Tuple[int, int, int, int]],
               transform: Optional[Callable[[np.ndarray], np.ndarray]] = None,
               verbose: int = 0,
               freq: float = 200,
               save_images: bool = False,
               img_name: str = "{self._n_loops:6d}_{t-self.t0:.6f}.tiff",
               save_folder: Optional[Union[str, Path]] = None,
               save_period: int = 1,
               save_backend: Optional[str] = None,
               image_generator: Optional[Callable[[float, float],
                                                  np.ndarray]] = None,
               labels: Optional[List[str]] = None,
               img_ref: Optional[np.ndarray] = None,
               kernel_file: Optional[Union[str, Path]] = None,
               iterations: int = 4,
               mul: float = 3,
               **kwargs) -> None:
    """Sets the args and initializes the camera.

    Args:
      camera: The name of the camera to control. See :ref:`Cameras` for an
        exhaustive list of available cameras.
      patches: A :obj:`list` containing the patches to track given as
        :obj:`tuple`. Each patch should contain in that order: the `y` coord of
        its origin, the `x` coord of its origin, its size along the `y` axis
        and its size along the `x` axis. Any number of patches can be given.
      transform: A function taking an image as an argument and returning a
        transformed image. The original image is discarded and only the
        transformed one is kept for processing, display and saving.
      verbose: The verbose level as an integer, between `0` and `3`. At level
        `0` no information is printed, and at level `3` so much information is
        printed that is slows the code down.
      freq: If given, the block will try to loop at this frequency. If it is
        lower than the framerate of the camera, frames will be dropped. This
        argument can be used for limiting the achieved framerate when the
        camera doesn't support framerate control.
      save_images: If :obj:`True`, the acquired images are saved on the
        computer during the test. Note that saving images uses CPU, so the
        achieved performance might drop when this feature is in use.
      img_name: If ``save_images`` is :obj:`True`, the template for naming the
        recorded images. It is evaluated as an `f-string`, and must contain the
        file extension at the end. For building the `f-string`, the
        ``self._n_loops`` attribute holds the loop number, and ``t-self.t0``
        holds the current timestamp.
      save_folder: If ``save_images`` is :obj:`True`, the directory to save
        images to. If it doesn't exist, it will be created. If not given, the
        images are saved in a folder named `Crappy_images` and created next to
        the file being run.
      save_period: If ``save_images`` is :obj:`True`, only one every this
        number of images will be saved.
      save_backend: The backend to use for saving the images. Should be one
        of :
        ::

          'sitk', 'pil', 'cv2'

        If not specified, SimpleITK will be used if available, then OpenCV as a
        second choice, and finally Pillow if none of the others was available.
      image_generator: A function taking two floats as arguments, and returning
        an image. It is only used for demonstration without camera in the
        examples, and isn't meant to be used in an actual test.
      labels: A :obj:`list` containing the labels to send to downstream blocks,
        carrying the information on the latest position of the patches. If
        not given, the default labels are :
        ::

          ['t(s)', 'meta', 'p0x', 'p0y', ..., pix, piy]

        with `i` the number of given patches.
      img_ref: The reference image to which all the acquired images will be
        compared for performing the correlation. If not given, the first
        acquired images will be used as the reference image.
      kernel_file: The path to a file containing the kernel modules for
        :mod:`pycuda` to use. If not given ,the default kernels of Crappy are
        used.
      iterations: The maximum number of iterations to run before returning the
        results. The results may be returned before if the residuals start
        increasing.
      mul: The scalar by which the direction will be multiplied before being
        added to the solution. If it's too high, the convergence will be fast
        but there's a risk that to go past the solution and to diverge. If it's
        too low, the convergence will be slower and require more iterations.
        `3` was found to be an acceptable value in most cases, but it is
        recommended to tune this value for each application so that the
        convergence is neither too slow nor too fast.
      **kwargs: Any additional argument to pass to the camera.
    """

    # Setting now the pycuda context to avoid setting it multiple times
    pycuda.driver.init()
    context = pycuda.tools.make_default_context()

    self._patches = patches
    self._img_ref = img_ref

    super().__init__(camera=camera,
                     transform=transform,
                     config=False,
                     display_images=False,
                     displayer_backend=None,
                     displayer_framerate=5,
                     software_trig_label=None,
                     verbose=bool(verbose),
                     freq=freq,
                     save_images=save_images,
                     img_name=img_name,
                     save_folder=save_folder,
                     save_period=save_period,
                     save_backend=save_backend,
                     image_generator=image_generator,
                     **kwargs)

    self._correls = [GPUCorrel_tool(context=context,
                                    verbose=verbose,
                                    levels=1,
                                    resampling_factor=2,
                                    kernel_file=kernel_file,
                                    iterations=iterations,
                                    mask=None,
                                    ref_img=img_ref,
                                    mul=mul,
                                    fields=['x', 'y']) for _ in self._patches]

    # Setting the labels
    if labels is None:
      self.labels = ['t(s)', 'meta'] + [elt
                                        for i, _ in enumerate(patches)
                                        for elt in [f'p{i}x', f'p{i}y']]
    else:
      self.labels = labels

    if 2 + 2 * len(patches) != len(self.labels):
      raise ValueError("The number of fields is inconsistent with the number "
                       "of labels !\nMake sure that the time and metadata "
                       "labels were given")

    # We can already set the sizes of the images as they are already known
    for correl, (_, __, h, w) in zip(self._correls, self._patches):
      correl.set_img_size((h, w))

  def prepare(self) -> None:
    """Opens the camera and sets the reference sub-image for each patch if a
    reference image was given."""

    super().prepare()

    if self._img_ref is not None:
      for correl, (oy, ox, h, w) in zip(self._correls, self._patches):
        correl.set_orig(self._img_ref[oy:oy + h, ox:ox + w].astype(np.float32))
        correl.prepare()

  def begin(self) -> None:
    """Acquires a first image and sets it as the reference image if no
    reference image was previously given."""

    if self._img_ref is None:
      _, img = self._camera.get_image()

      for correl, (oy, ox, h, w) in zip(self._correls, self._patches):
        correl.set_orig(img[oy:oy + h, ox:ox + w].astype(np.float32))
        correl.prepare()

  def finish(self) -> None:
    """Closes the correlation objects as well as the camera."""

    for correl in self._correls:
      correl.clean()

    super().finish()

  def _additional_loop(self, meta: Dict[str, Any], img: np.ndarray) -> None:
    """Gets the updated positions of the patches, and sends it to the
    downstream blocks."""

    out = [meta['t(s)'], meta]
    for correl, (oy, ox, h, w) in zip(self._correls, self._patches):
      out.extend(correl.get_disp(img[oy:oy + h,
                                 ox:ox + w].astype(np.float32)).tolist())

    self.send(out)



# coding: utf-8

"""
Demonstrates the use of the Drawing block.

No hardware required.
Requires the cv2 module to be installed.
"""

from typing import List
import crappy


class TestBlock(crappy.Block):
  """A simple block for demonstrating the use of the Drawing block.

  Simply updates the values on the Drawing at each loop.
  """

  def __init__(self, labels: List[str]) -> None:
    """Sets the args and initializes the parent class."""

    super().__init__()
    self.freq = 5

    self._labels = labels
    self._loops = 0

  def loop(self) -> None:
    """At each loop, returns the updated dict of values to display."""

    ret = {}
    for i, label in enumerate(self._labels):
      ret[label] = self._loops + 15 * i
    self.send(ret)
    self._loops += 1


if __name__ == "__main__":

  # The path to the background image
  img_path = crappy.resources.paths['pad']

  # The coordinates of the thermocouples on the drawing
  coord = [(185, 430), (145, 320), (105, 220), (720, 370), (720, 250),
           (720, 125), (1220, 410), (1260, 320), (1300, 230)]

  # The dict containing the information on the elements to display on the
  # Drawing. Here we add 'dot_text' elements.
  options = [{'type': 'dot_text',
              'coord': coord[i],
              'text': f'T{i + 1} = %.1f',
              'label': f'T{i + 1}'} for i in range(len(coord))]

  # The block that will update the values of the 'dot_text' elements on the
  # Drawing. We simply give it a list containing the label of each element.
  update_block = TestBlock([d['label'] for d in options])

  # Now adding a 'time' element to the Drawing. The 'text' and 'label' keys
  # are not mandatory and do not affect the Drawing.
  options.append({'type': 'time',
                  'coord': (80, 1000),
                  'text': '',
                  'label': ''})

  # The block containing the Drawing
  drawing = crappy.blocks.Drawing(image=img_path,
                                  draw=options,
                                  color_range=(20, 300),
                                  title="Temperatures")

  # Linking and starting the blocks
  crappy.link(update_block, drawing)
  crappy.start()



# coding: utf-8

"""
Example demonstrating the use of the generator with steps.

In this example, we would like to reach different levels of strain and relax
the sample (return to `F=0`) between each strain level.

No required hardware.
"""

import crappy

if __name__ == "__main__":
  speed = 5 / 60  # mm/sec

  path = []  # We will put in this list all the paths to be followed

  # We will loop over the values we would like to reach
  # And add two paths for each loop: one for loading and one for unloading
  for exx in [.25, .5, .75, 1., 1.5, 2]:
    path.append({'type': 'constant',
                 'value': speed,
                 'condition': 'Exx(%)>{}'.format(exx)})  # Go up to this level
    path.append({'type': 'constant',
                 'value': -speed,
                 'condition': 'F(N)<0'})  # Go down to F=0N

  # Now we can simply give our list of paths to the generator
  generator = crappy.blocks.Generator(path=path)

  # This block will simulate a tensile testing machine
  machine = crappy.blocks.Fake_machine()
  # We must link the generator to the machine to give the command to
  # the machine
  crappy.link(generator, machine)
  # But also the machine to the generator because we added conditions on force
  # and strain, so the generator needs these values coming out of the machine
  # Remember : links are one way only !
  crappy.link(machine, generator)

  # Let's add two graphs to visualise in real time
  graph_def = crappy.blocks.Grapher(('t(s)', 'Exx(%)'))
  crappy.link(machine, graph_def)

  graph_f = crappy.blocks.Grapher(('t(s)', 'F(N)'))
  crappy.link(machine, graph_f)

  # And start the experiment
  crappy.start()



# coding: utf-8

from time import time
from typing import List
from .inout import InOut
from .._global import OptionalModule

try:
  import serial
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("pyserial")


class Agilent34420a(InOut):
  """Sensor class for Agilent34420A devices.

  This class contains method to measure values of resistance or voltage on
  Agilent34420A devices.

  Note:
    May work for other devices too, but not tested.

    If you have issues with this class returning a lot of `'bad serial'`, make
    sure you have the last version of :mod:`serial`.
  """

  def __init__(self,
               mode: bytes = b"VOLT",
               device: str = '/dev/ttyUSB0',
               baudrate: int = 9600,
               timeout: float = 1) -> None:
    """Sets the args and initializes parent class.

    Args:
      mode: Desired value to measure. Should be either `b'VOLT'` or `b'RES'`.
      device: Path to the device to open, as a :obj:`str`.
      baudrate: Desired baudrate for serial communication.
      timeout: Timeout for the serial connection, as a :obj:`float`.
    """

    super().__init__()

    self._device = device
    self._baudrate = baudrate
    self._timeout = timeout
    self._mode = mode

    self._ser = None

  def open(self) -> None:
    """Opens the serial connection, resets the Agilent and configures it to the
    desired mode."""

    self._ser = serial.Serial(port=self._device, baudrate=self._baudrate,
                              timeout=self._timeout)
    self._ser.write(b"*RST;*CLS;*OPC?\n")
    self._ser.write(b"SENS:FUNC \"" + self._mode + b"\";  \n")
    self._ser.write(b"SENS:" + self._mode + b":NPLC 2  \n")
    self._ser.write(b"SYST:REM\n")

  def get_data(self) -> List[float]:
    """Asks the Agilent to acquire a reading and returns it, except if an error
    occurs in which case `0` is returned."""

    self._ser.write(b"READ?  \n")
    t = time()
    try:
      return [t, float(self._ser.readline())]
    except (serial.SerialException, ValueError):
      self._ser.flush()
      return [t, 0]

  def close(self) -> None:
    """Closes the serial port."""

    self._ser.close()



# coding: utf-8

from .inout import InOut, MetaIO

from .ads1115 import Ads1115
from .agilent34420A import Agilent34420a
from .comedi import Comedi
from .fake_inout import Fake_inout
from .gpio_pwm import Gpio_pwm
from .gpio_switch import Gpio_switch
from .gsm import Gsm
from .kollmorgen import Koll
from .labjackT7 import Labjack_t7
from .labjackUE9 import Labjack_ue9
from .mcp9600 import Mcp9600
from .mprls import Mprls
from .nau7802 import Nau7802
from .ni_daqmx import Nidaqmx
from .opsens import Opsens
from .piJuice import Pijuice
from .spectrum import Spectrum
from .t7Streamer import T7_streamer
from .waveshare_ad_da import Waveshare_ad_da
from .waveshare_ad_da_ft232h import Waveshare_ad_da_ft232h
from .waveshare_high_precision import Waveshare_high_precision

# Win specific
from .daqmx import Daqmx

# All the inout objects
inout_dict = MetaIO.classes



# coding: utf-8

"""
A basic example showing how to use the GPUCorrel block.

Required hardware:
  - Any camera
  - A CUDA compatible GPU
"""

import crappy

if __name__ == "__main__":
  graph = crappy.blocks.Grapher(('t(s)', 'x'), ('t(s)', 'y'), ('t(s)', 'r'),
                                length=50)

  correl = crappy.blocks.GPUCorrel(camera="Webcam",
                                   fields=['x', 'y', 'r'])  # Rigid body

  crappy.link(correl, graph)

  crappy.start()



# coding:utf-8

import warnings
from math import ceil
import numpy as np
from pkg_resources import resource_filename
from typing import Any, Tuple, Optional, Union, List
from pathlib import Path
from itertools import chain

from .fields import get_field
from .._global import OptionalModule
try:
  import pycuda.tools
  import pycuda.driver
  import pycuda.compiler
  import pycuda.gpuarray as gpuarray
  import pycuda.reduction
except (ImportError, ModuleNotFoundError):
  pycuda = OptionalModule("pycuda",
                          "PyCUDA and CUDA are necessary to use GPUCorrel")
  gpuarray = OptionalModule("pycuda",
                            "PyCUDA and CUDA are necessary to use GPUCorrel")


def interp_nearest(arr: np.ndarray, ny: int, nx: int) -> np.ndarray:
  """Reshapes an input array to the specified dimension using the nearest
  interpolation, using only :mod:`numpy` methods.

  Args:
    arr: The array to interpolate.
    ny: The new dimension along the `y` axis.
    nx: The new dimension along the `y` axis.

  Returns:
    A reshaped version of the input array obtained with the nearest
    interpolation.
  """

  # If the shape is already fine, nothing more to do
  if arr.shape == (ny, nx):
    return arr

  y, x = arr.shape
  rx, ry = x / nx, y / ny
  out = np.empty((ny, nx), dtype=np.float32)

  # Filling individually each value of the returned array
  for j in range(ny):
    for i in range(nx):
      out[j, i] = arr[int(ry * j + .5), int(rx * i + .5)]

  return out


class CorrelStage:
  """Represents a stage of the pyramid used by the :ref:`GPUCorrel` tool for
  performing GPU correlation.

  This class actually performs the GPU computation, while the calling classes
  only manage the data.
  """

  def __init__(self,
               img_size: Tuple[int, int],
               verbose: int = 0,
               iterations: int = 5,
               mul: float = 3,
               n_fields: Optional[int] = None,
               kernel_file: Optional[Union[Path, str]] = None) -> None:
    """Sets the args and instantiates the :mod:`pycuda` objects.

    Args:
      img_size: The shape of the images to process. It is given beforehand so
        that the memory can be allocated before the test starts.
      verbose: The verbose level as an integer, between `0` and `3`. At level
        `0` no information is printed, and at level `3` so much information is
        printed that is slows the code down.
      iterations: The maximum number of iterations to run before returning the
        results. The results may be returned before if the residuals start
        increasing.
      mul: The scalar by which the direction will be multiplied before being
        added to the solution. If it's too high, the convergence will be fast
        but there's a risk that to go past the solution and to diverge. If it's
        too low, the convergence will be slower and require more iterations.
        `3` was found to be an acceptable value in most cases, but it is
        recommended to tune this value for each application so that the
        convergence is neither too slow nor too fast.
      n_fields: The number of fields to project the displacement on.
      kernel_file: The path to the file containing the kernels to use for the
        correlation. Can be a :obj:`pathlib.Path` object or a :obj:`str`.
    """

    # Setting the args
    self._verbose = verbose
    self._iterations = iterations
    self._mul = mul
    self._n_fields = n_fields
    self._ready = False
    self._height, self._width = img_size
    self._debug(2, f"Initializing with resolution {img_size}")

    # These attributes will be set later
    self._r_x, self._r_y = None, None
    self._mask_array = None
    self._array = None
    self._array_d = None
    self._fields = False
    self._r_grid = None
    self._r_block = None
    self._dev_r_out = None
    self.res = None

    # Setting the grid and the block for the kernel
    self._grid = (int(ceil(self._width / 32)), int(ceil(self._height / 32)))
    self._block = (int(ceil(self._width / self._grid[0])),
                   int(ceil(self._height / self._grid[1])), 1)
    self._debug(3, f"Default grid: {self._grid}, block: {self._block}")

    # dev_g stores the G arrays (to compute the research direction)
    self._dev_g = [gpuarray.empty(img_size, np.float32)
                   for _ in range(self._n_fields)]
    # dev_fields_x/y store the fields value along X and Y
    self._dev_fields_x = [gpuarray.empty(img_size, np.float32)
                          for _ in range(self._n_fields)]
    self._dev_fields_y = [gpuarray.empty(img_size, np.float32)
                          for _ in range(self._n_fields)]
    # dev_h Stores the Hessian matrix
    self._dev_h = np.zeros((self._n_fields, self._n_fields), np.float32)
    # And dev_h_i stores its invert
    self.dev_h_i = gpuarray.empty((self._n_fields, self._n_fields), np.float32)
    # dev_out is written with the difference of the images
    self._dev_out = gpuarray.empty((self._height, self._width), np.float32)
    # dev_x stores the value of the parameters (what is actually computed)
    self.dev_x = gpuarray.empty(self._n_fields, np.float32)
    # dev_vec stores the research direction
    self._dev_vec = gpuarray.empty(self._n_fields, np.float32)
    # dev_orig stores the original image on the device
    self.dev_orig = gpuarray.empty(img_size, np.float32)
    # dev_grad_x store the gradient along X of the original image on the device
    self._dev_grad_x = gpuarray.empty(img_size, np.float32)
    # And along Y
    self._dev_grad_y = gpuarray.empty(img_size, np.float32)

    # Opening the kernel file and compiling the module
    if kernel_file is None:
      self._debug(2, "Kernel file not specified")
      kernel_file = resource_filename('crappy', 'tool/kernels.cu')
    with open(kernel_file, "r") as file:
      self._debug(3, "Sourcing module")
      mod = pycuda.compiler.SourceModule(file.read() % (self._width,
                                                        self._height,
                                                        self._n_fields))

    # Extracting individual functions from the data of the kernel file
    self._resample_orig_krnl = mod.get_function('resampleO')
    self._resample_krnl = mod.get_function('resample')
    self._gradient_krnl = mod.get_function('gradient')
    self._make_g_krnl = mod.get_function('makeG')
    self._make_diff = mod.get_function('makeDiff')
    self._dot_krnl = mod.get_function('myDot')
    self._add_krnl = mod.get_function('kadd')
    self._mul_red_krnl = pycuda.reduction.ReductionKernel(
      np.float32, neutral="0", reduce_expr="a+b", map_expr="x[i]*y[i]",
      arguments="float *x, float *y")
    self._least_square = pycuda.reduction.ReductionKernel(
      np.float32, neutral="0", reduce_expr="a+b", map_expr="x[i]*x[i]",
      arguments="float *x")

    # Getting the texture references
    self._tex = mod.get_texref('tex')
    self._tex_d = mod.get_texref('tex_d')
    self._tex_mask = mod.get_texref('texMask')

    # All textures use normalized coordinates except for the mask
    for t in (self._tex, self._tex_d):
      t.set_flags(pycuda.driver.TRSF_NORMALIZED_COORDINATES)
    for t in (self._tex, self._tex_d, self._tex_mask):
      t.set_filter_mode(pycuda.driver.filter_mode.LINEAR)
      t.set_address_mode(0, pycuda.driver.address_mode.BORDER)
      t.set_address_mode(1, pycuda.driver.address_mode.BORDER)

    # Preparing kernels for less overhead when called
    self._resample_orig_krnl.prepare("Pii", texrefs=[self._tex])
    self._resample_krnl.prepare("Pii", texrefs=[self._tex_d])
    self._gradient_krnl.prepare("PP", texrefs=[self._tex])
    self._make_diff.prepare("PPPP",
                            texrefs=[self._tex, self._tex_d, self._tex_mask])
    self._add_krnl.prepare("PfP")

  def set_orig(self, img) -> None:
    """Sets the original image from a given CPU or GPU array"""

    if isinstance(img, np.ndarray):
      self._debug(3, "Setting original image from ndarray")
      self.dev_orig.set(img)

    elif isinstance(img, gpuarray.GPUArray):
      self._debug(3, "Setting original image from GPUArray")
      self.dev_orig = img

    else:
      raise ValueError("Error ! Unknown type of data given to set_orig()")

    self.update_orig()

  def update_orig(self) -> None:
    """Updates the original image."""

    self._debug(3, "Updating original image")

    self._array = pycuda.driver.gpuarray_to_array(self.dev_orig, 'C')
    self._tex.set_array(self._array)
    self._compute_gradients()
    self._ready = False

  def prepare(self) -> None:
    """Computes all the necessary tables to perform correlation."""

    # Sets the mask array if none was specified
    if self._mask_array is None:
      self._debug(2, "No mask set when preparing, using a basic one, with a "
                     "border of 5% the dimension")
      mask = np.zeros((self._height, self._width), np.float32)
      mask[self._height // 20: -self._height // 20,
           self._width // 20: -self._width // 20] = 1
      self.set_mask(mask)

    # Only necessary to prepare if not already ready
    if not self._ready:
      # Checking that everything's set for preparing
      if self._array is None:
        self._debug(1, "Tried to prepare but original texture is not set !")
      elif not self._fields:
        self._debug(1, "Tried to prepare but fields are not set !")

      # Actually computing the tables
      else:
        self._make_g()
        self._make_h()
        self._ready = True
        self._debug(3, "Ready!")

    else:
      self._debug(1, "Tried to prepare when unnecessary, doing nothing...")

  def resample_orig(self, new_y: int, new_x: int, dev_out) -> None:
    """Resamples the original image with a new dimension."""

    # New grid and block to be used
    grid = (int(ceil(new_x / 32)), int(ceil(new_y / 32)))
    block = (int(ceil(new_x / grid[0])), int(ceil(new_y / grid[1])), 1)
    self._debug(3, f"Resampling Orig texture, grid: {grid}, block: {block}")

    # Now resampling
    self._resample_orig_krnl.prepared_call(self._grid, self._block,
                                           dev_out.gpudata,
                                           np.int32(new_x), np.int32(new_y))
    self._debug(3, f"Resampled original texture to {dev_out.shape}")

  def resample_d(self, new_y: int, new_x: int) -> Any:
    """Resamples the texture with a new dimension and returns it in a GPU
    array."""

    if (self._r_x, self._r_y) != (np.int32(new_x), np.int32(new_y)):
      self._r_grid = (int(ceil(new_x / 32)), int(ceil(new_y / 32)))
      self._r_block = (int(ceil(new_x / self._r_grid[0])),
                       int(ceil(new_y / self._r_grid[1])), 1)
      self._r_x, self._r_y = np.int32(new_x), np.int32(new_y)
      self._dev_r_out = gpuarray.empty((new_y, new_x), np.float32)

    self._debug(3, f"Resampling img_d texture to {new_y, new_x}, "
                   f"grid: {self._r_grid}, block:{self._r_block}")

    self._resample_krnl.prepared_call(self._r_grid, self._r_block,
                                      self._dev_r_out.gpudata,
                                      self._r_x, self._r_y)
    return self._dev_r_out

  def set_fields(self, fields_x, fields_y) -> None:
    """Sets the fields on which hto project the displacement."""

    self._debug(2, "Setting fields")

    if isinstance(fields_x, np.ndarray):
      self._dev_fields_x.set(fields_x)
      self._dev_fields_y.set(fields_y)

    elif isinstance(fields_x, gpuarray.GPUArray):
      self._dev_fields_x = fields_x
      self._dev_fields_y = fields_y

    self._fields = True

  def set_image(self, img_d) -> None:
    """Set the current image, to be compared with the reference image."""

    if isinstance(img_d, np.ndarray):
      self._debug(3, "Creating texture from numpy array")
      self._array_d = pycuda.driver.matrix_to_array(img_d, "C")

    elif isinstance(img_d, gpuarray.GPUArray):
      self._debug(3, "Creating texture from gpuarray")
      self._array_d = pycuda.driver.gpuarray_to_array(img_d, "C")

    else:
      self._debug(0, "Error ! Unknown type of data given to .set_image()")
      raise ValueError

    self._tex_d.set_array(self._array_d)
    self.dev_x.set(np.zeros(self._n_fields, dtype=np.float32))

  def set_mask(self, mask) -> None:
    """Sets the mask to use for weighting the images."""

    self._debug(3, "Setting the mask")

    if not mask.dtype == np.float32:
      self._debug(2, "Converting the mask to float32")
      mask = mask.astype(np.float32)

    if isinstance(mask, np.ndarray):
      self.maskArray = pycuda.driver.matrix_to_array(mask, 'C')

    elif isinstance(mask, gpuarray.GPUArray):
      self.maskArray = pycuda.driver.gpuarray_to_array(mask, 'C')

    else:
      raise ValueError("Error! Mask data type not understood")

    self._tex_mask.set_array(self.maskArray)

  def set_disp(self, x) -> None:
    """Sets the displacement fields computed from the previous stages."""

    if isinstance(x, gpuarray.GPUArray):
      self.dev_x = x

    elif isinstance(x, np.ndarray):
      self.dev_x.set(x)

    else:
      raise ValueError("Error! Unknown type of data given to "
                       "CorrelStage.set_disp")

  def get_disp(self, img_d=None) -> Any:
    """Projects the displacement on the base fields, and returns the result."""

    self._debug(3, "Calling main routine")

    if not self._ready:
      self._debug(2, "Wasn't ready ! Preparing...")
      self.prepare()

    if img_d is not None:
      self.set_image(img_d)

    if self._array_d is None:
      raise ValueError("Did not set the image, use set_image() before calling "
                       "get_disp !")

    self._debug(3, "Computing first diff table")

    self._make_diff.prepared_call(self._grid, self._block,
                                  self._dev_out.gpudata,
                                  self.dev_x.gpudata,
                                  self._dev_fields_x.gpudata,
                                  self._dev_fields_y.gpudata)

    self.res = self._least_square(self._dev_out).get()
    self._debug(3, f"res: {self.res / 1e6}")

    for i in range(self._iterations):
      self._debug(3, f"Iteration {i}")

      # Computing the direction of the gradient of each parameter
      for j in range(self._n_fields):
        self._dev_vec[j] = self._mul_red_krnl(self._dev_g[j], self._dev_out)

      # Newton method: the gradient vector is multiplied by the pre-inverted
      # Hessian, self._dev_vec now contains the actual research direction
      self._dot_krnl(self.dev_h_i, self._dev_vec,
                     grid=(1, 1), block=(self._n_fields, 1, 1))

      # This line simply adds mul times the research direction to self._dev_x
      # with a really simple kernel (does self._dev_x += mul * self._dev_vec)
      self._add_krnl.prepared_call((1, 1), (self._n_fields, 1, 1),
                                   self.dev_x.gpudata, self._mul,
                                   self._dev_vec.gpudata)

      # Simple way to prevent unnecessary copy of data
      if self._verbose >= 3:
        self._debug(3, f"Direction: {self._dev_vec.get()}")
        self._debug(3, f"New X: {self.dev_x.get()}")

      # Getting the new residuals
      self._make_diff.prepared_call(self._grid, self._block,
                                    self._dev_out.gpudata,
                                    self.dev_x.gpudata,
                                    self._dev_fields_x.gpudata,
                                    self._dev_fields_y.gpudata)

      prev_res = self.res
      self.res = self._least_square(self._dev_out).get()

      # Handling the case when the residuals start increasing
      if self.res >= prev_res:
        self._debug(3, f"Diverting from the solution "
                       f"new res={self.res / 1e6} >= {prev_res / 1e6}!")

        self._add_krnl.prepared_call((1, 1), (self._n_fields, 1, 1),
                                     self.dev_x.gpudata,
                                     -self._mul,
                                     self._dev_vec.gpudata)
        self.res = prev_res
        self._debug(3, f"Undone: X={self.dev_x.get()}")
        break

      self._debug(3, f"res: {self.res / 1e6}")

    return self.dev_x.get()

  def get_data_display(self) -> np.ndarray:
    """Returns the necessary data for displaying the difference between the
    reference and current image."""

    return (self._dev_out.get() + 128).astype(np.uint8)

  def _make_g(self) -> None:
    """Computes the tables."""

    for i in range(self._n_fields):
      self._make_g_krnl(self._dev_g[i].gpudata, self._dev_grad_x.gpudata,
                        self._dev_grad_y.gpudata,
                        self._dev_fields_x[i], self._dev_fields_y[i],
                        block=self._block, grid=self._grid)

  def _make_h(self) -> None:
    """Computes the tables."""

    for i in range(self._n_fields):
      for j in range(i + 1):
        self._dev_h[i, j] = self._mul_red_krnl(self._dev_g[i],
                                               self._dev_g[j]).get()
        if i != j:
          self._dev_h[j, i] = self._dev_h[i, j]

    self._debug(3, f"Hessian: {self._dev_h}")
    self.dev_h_i.set(np.linalg.inv(self._dev_h))

    if self._verbose >= 3:
      self._debug(3, f"Inverted Hessian: {self.dev_h_i.get()}")

  def _debug(self, level: int, msg: str) -> None:
    """Prints the provided debug message only if its debug level is lower than
    or equal to the verbose level."""

    if level <= self._verbose:
      print(f"[ GPUCorrel level {level} debug] {msg}")

  def _compute_gradients(self) -> None:
    """Wrapper to call the gradient kernel."""

    self._gradient_krnl.prepared_call(self._grid, self._block,
                                      self._dev_grad_x.gpudata,
                                      self._dev_grad_y.gpudata)


class GPUCorrel:
  """This tool performs GPU correlation between two images in order to identify
  fields of rigid body displacements.

  This class  is meant to be efficient enough to run in real-time. It relies on
  :class:`CorrelStage` to perform correlation on different scales. It mainly
  takes a list of base fields and a reference image as inputs, and project the
  displacement between the current image and the reference one on the base of
  fields. The optimal fit is achieved by lowering the residuals with a
  least-squares method.

  The projection on the base is performed sequentially, using the results
  obtained at stages with low resolution to initialize the computation on
  on stages with higher resolution. A newton method is used to converge towards
  an optimal solution.
  """

  context = None

  def __init__(self,
               context: Optional[Any] = None,
               verbose: int = 0,
               levels: int = 5,
               resampling_factor: float = 2,
               kernel_file: Optional[Union[str, Path]] = None,
               iterations: int = 4,
               fields: Optional[List[str]] = None,
               ref_img: Optional[np.ndarray] = None,
               mask: Optional[np.ndarray] = None,
               mul: float = 3) -> None:
    """Sets the args and a few parameters of :mod:`pycuda`.

    Args:
      context: Optionally, the :mod:`pycuda` context to use. If not specified,
        a new context is instantiated.
      verbose: The verbose level as an integer, between `0` and `3`. At level
        `0` no information is printed, and at level `3` so much information is
        printed that is slows the code down.
      levels: Number of levels of the pyramid. More levels may help converging
        on images with large strain, but may fail on images that don't contain
        low spatial frequency. Fewer levels mean that the program runs faster.
      resampling_factor: the factor by which the resolution is divided between
        each stage of the pyramid. A low factor ensures coherence between the
        stages, but is more computationally intensive. A high factor allows
        reaching a finer detail level, but may lead to a coherence loss between
        the stages.
      kernel_file: The path to the file containing the kernels to use for the
        correlation. Can be a :obj:`pathlib.Path` object or a :obj:`str`.
      iterations: The maximum number of iterations to run before returning the
        results. The results may be returned before if the residuals start
        increasing.
      fields: A :obj:`list` of :obj:`str` representing the base of fields on
        which the image will be projected during correlation. The possible
        fields are :
        ::

          'x', 'y', 'r', 'exx', 'eyy', 'exy', 'eyx', 'exy2', 'z'

      ref_img: the reference image, as a 2D :mod:`numpy` array with `dtype`
        `float32`. It can either be given at ``__init__``, or set later with
        :meth:`set_orig`.
      mask: The mask used for weighting the region of interest on the image. It
        is generally used to prevent unexpected behavior on the border of the
        image.
      mul: The scalar by which the direction will be multiplied before being
        added to the solution. If it's too high, the convergence will be fast
        but there's a risk that to go past the solution and to diverge. If it's
        too low, the convergence will be slower and require more iterations.
        `3` was found to be an acceptable value in most cases, but it is
        recommended to tune this value for each application so that the
        convergence is neither too slow nor too fast.
    """

    self._context = context

    self._verbose = verbose
    self._debug(3, "You set the verbose level to the maximum.\nIt may help "
                   "finding bugs or tracking errors but it may also impact the"
                   " program performance as it will print A LOT of output and "
                   "add GPU->CPU copies only to print information.\nIf it is "
                   "not desired, consider lowering the verbosity: 1 or 2 is a "
                   "reasonable choice, 0 won't show anything except errors.")

    # Setting the args
    self._levels = levels
    self._resampling_factor = resampling_factor
    self._iterations = iterations
    self._mul = mul
    self._fields = fields
    self._n_fields = len(fields)
    self._mask = mask
    self._ref_img = ref_img
    self._loops = 0
    self._last = None

    # These attributes will be set later
    self._heights, self._widths = None, None
    self._stages = None
    self._tex_fx = None
    self._tex_fy = None
    self._resample = None

    self._debug(1, f"Initializing... levels: {levels}, verbosity: {verbose}")

    # Getting the kernel file
    if kernel_file is None:
      self._debug(3, "Kernel file not specified, using the default one of "
                     "crappy")
      self._kernel_file = resource_filename('crappy', 'tool/kernels.cu')
    else:
      self._kernel_file = kernel_file
    self._debug(3, f"Kernel file:{self._kernel_file}")

  def set_img_size(self, img_size: Tuple[int, int]) -> None:
    """Sets the image shape, and calls the methods that need this information
    for running."""

    self._debug(1, f"Setting master resolution: {img_size},")

    # Initializing the pycuda context
    if self._context is not None:
      GPUCorrel.context = self._context
    else:
      pycuda.driver.init()
      GPUCorrel.context = pycuda.tools.make_default_context()

    src_txt = """
        texture<float, cudaTextureType2D, cudaReadModeElementType> texFx{0};
        texture<float, cudaTextureType2D, cudaReadModeElementType> texFy{0};
        __global__ void resample{0}(float* outX, float* outY, int x, int y)
        {{
          int idx = blockIdx.x*blockDim.x+threadIdx.x;
          int idy = blockIdx.y*blockDim.y+threadIdx.y;
          if(idx < x && idy < y)
          {{
            outX[idy*x+idx] = tex2D(texFx{0},(float)idx/x, (float)idy/y);
            outY[idy*x+idx] = tex2D(texFy{0},(float)idx/x, (float)idy/y);
          }}
        }}
        """

    # Setting the textures for a faster resampling
    src = ''.join([src_txt.format(i) for i in range(self._n_fields)])
    source_module = pycuda.compiler.SourceModule(src)

    self._tex_fx = [source_module.get_texref(f"texFx{i}")
                    for i in range(self._n_fields)]
    self._tex_fy = [source_module.get_texref(f"texFy{i}")
                    for i in range(self._n_fields)]
    self._resample = [source_module.get_function(f"resample{i}")
                      for i in range(self._n_fields)]

    for tex_fx, tex_fy, resample in zip(self._tex_fx, self._tex_fy,
                                        self._resample):
      resample.prepare('PPii', texrefs=[tex_fx, tex_fy])

    for tex in chain(self._tex_fx, self._tex_fy):
      tex.set_flags(pycuda.driver.TRSF_NORMALIZED_COORDINATES)
      tex.set_filter_mode(pycuda.driver.filter_mode.LINEAR)
      tex.set_address_mode(0, pycuda.driver.address_mode.BORDER)
      tex.set_address_mode(1, pycuda.driver.address_mode.BORDER)

    # Setting the dimensions for each stage
    height, width, *_ = img_size
    self._heights = [round(height / (self._resampling_factor ** i))
                     for i in range(self._levels)]
    self._widths = [round(width / (self._resampling_factor ** i))
                    for i in range(self._levels)]

    # Initializing all the stages
    self._stages = [CorrelStage(img_size=(height, width),
                                verbose=self._verbose,
                                n_fields=self._n_fields,
                                iterations=self._iterations,
                                mul=self._mul,
                                kernel_file=self._kernel_file)
                    for i, (height, width) in enumerate(zip(self._heights,
                                                            self._widths))]

    # Now that the stages exist, setting the reference image, fields, and mask
    if self._ref_img is not None:
      self.set_orig(self._ref_img)

    if self._fields is not None:
      self._set_fields(self._fields)

    if self._mask is not None:
      self._set_mask(self._mask)

  def set_orig(self, img: np.ndarray) -> None:
    """Sets the reference image, to which the following images will be
    compared."""

    self._debug(2, "Updating the original image")

    # Casting to float32 if needed
    if img.dtype != np.float32:
      warnings.warn("Correl() takes arrays with dtype np.float32 to allow GPU "
                    "computing (got {}). Converting to float32."
                    .format(img.dtype), RuntimeWarning)
      img = img.astype(np.float32)

    # Setting the reference image for all stages
    self._stages[0].set_orig(img)
    for prev_stage, stage, height, width in zip(self._stages[:-1],
                                                self._stages[1:],
                                                self._heights, self._widths):
      prev_stage.resample_orig(height, width, stage.dev_orig)
      stage.update_orig()

  def prepare(self) -> None:
    """Prepares all the stages before starting the test."""

    for stage in self._stages:
      stage.prepare()

    self._debug(2, "Ready !")

  def get_disp(self, img_d: Optional[np.ndarray] = None) -> Any:
    """To get the displacement.

    This will perform the correlation routine on each stage, initializing with
    the previous values every time it will return the computed parameters
    as a list.
    """

    if img_d is not None:
      self._set_image(img_d)

    if self._last is not None:
      disp = self._last / (self._resampling_factor ** self._levels)
    else:
      disp = np.array([0] * self._n_fields, dtype=np.float32)

    for stage in reversed(self._stages):
      disp *= self._resampling_factor
      stage.set_disp(disp)
      disp = stage.get_disp()

    self._loops += 1
    if not self._loops % 10:
      self._debug(2,
                  f"Loop {self._loops}, values: {self._stages[0].dev_x.get()}"
                  f", res: {self._stages[0].res / 1e6}")

    return disp

  def get_data_display(self) -> np.ndarray:
    """Returns the necessary data for displaying the difference between the
    reference and current image."""

    return self._stages[0].get_data_display()

  def get_res(self, lvl: int = 0):
    """Returns the last residual of the specified level."""

    return self._stages[lvl].res

  @staticmethod
  def clean():
    """Needs to be called at the end, to destroy the :mod:`pycuda` context
    properly."""

    GPUCorrel.context.pop()

  def _get_fields(self,
                  y: Optional[int] = None,
                  x: Optional[int] = None) -> (Any, Any):
    """Returns the fields, resampled to size `(y,x)`."""

    if x is None or y is None:
      y, x = self._heights[0], self._widths[0]

    out_x = gpuarray.empty((self._n_fields, y, x), np.float32)
    out_y = gpuarray.empty((self._n_fields, y, x), np.float32)
    grid = (int(ceil(x / 32)), int(ceil(y / 32)))
    block = (int(ceil(x / grid[0])), int(ceil(y / grid[1])), 1)

    for i, resample in enumerate(self._resample):
      resample.prepared_call(grid, block,
                             out_x[i, :, :].gpudata,
                             out_y[i, :, :].gpudata,
                             np.int32(x), np.int32(y))

    return out_x, out_y

  def _debug(self, level: int, msg: str) -> None:
    """Prints the provided debug message only if its debug level is lower than
    or equal to the verbose level."""

    if level <= self._verbose:
      print(f"[ GPUCorrel level {level} debug] {msg}")

  def _set_fields(self, fields: List[str]) -> None:
    """Computes the fields based on the provided field strings, and sets them
    for each stage."""

    for field_str, tex_fx, tex_fy in zip(fields, self._tex_fx, self._tex_fy):

      # Getting the fields as numpy arrays
      field_x, field_y = get_field(field_str, self._heights[0],
                                   self._widths[0])

      tex_fx.set_array(pycuda.driver.matrix_to_array(field_x, 'C'))
      tex_fy.set_array(pycuda.driver.matrix_to_array(field_y, 'C'))

    # Setting the fields for each stage
    for stage, height, width in zip(self._stages, self._heights, self._widths):
      stage.set_fields(*self._get_fields(height, width))

  def _set_image(self, img_d: np.ndarray) -> None:
    """Sets the current image for all the stages, to be compared with the
    reference image."""

    # Casting to float32 if needed
    if img_d.dtype != np.float32:
      warnings.warn(f"Correl() takes arrays with dtype np.float32 to allow GPU"
                    f" computing (got {img_d.dtype}). Converting to float32.",
                    RuntimeWarning)
      img_d = img_d.astype(np.float32)

    # Setting the current image for all stages
    self._stages[0].set_image(img_d)
    for prev_stage, stage, height, width in zip(self._stages[:-1],
                                                self._stages[1:],
                                                self._heights[1:],
                                                self._widths[1:]):
      stage.set_image(prev_stage.resample_d(height, width))

  def _set_mask(self, mask: np.ndarray) -> None:
    """Sets for each field the mask for weighting the images to process."""

    for stage, height, width in zip(self._stages, self._heights, self._widths):
      stage.set_mask(interp_nearest(mask, height, width))



# coding: utf-8

from time import time
from typing import Optional, List
from dataclasses import dataclass

from .inout import InOut
from ..tool import comedi_bind as comedi


@dataclass
class _Channel:
  """This class is a simple structure holding all the attributes a Comedi
  channel can have."""

  num: int
  range_num: int = 0
  gain: float = 1
  offset: float = 0
  make_zero: bool = False
  max_data: int = 0
  range_ds: int = 0


class Comedi(InOut):
  """This class can control acquisition boards relying on the Comedi driver.

  It can read data from ADCs on input channels, and set voltages of DACs on
  output channels. Each channel can be tuned independently in terms of range,
  gan, offset, and for input channels it's possible to decide whether they
  should be offset to `0` at the beginning of the test.
  """

  def __init__(self,
               device: str = '/dev/comedi0',
               subdevice: int = 0,
               channels: Optional[List[int]] = None,
               range_num: Optional[List[int]] = None,
               gain: Optional[List[float]] = None,
               offset: Optional[List[float]] = None,
               make_zero: Optional[List[bool]] = None,
               out_subdevice: int = 1,
               out_channels: Optional[List[int]] = None,
               out_range_num: Optional[List[int]] = None,
               out_gain: Optional[List[float]] = None,
               out_offset: Optional[List[float]] = None) -> None:
    """Sets the args and initializes the parent class.

    Args:
      device: The address of the device, as a :obj:`str`.
      subdevice: The id of the subdevice to use for input channels, as an
        :obj:`int`.
      channels: A :obj:`list` containing the indexes of the channels to use as
        inputs, given as :obj:`int`.
      range_num: A :obj:`list` containing for each input channel the index of
        the range to set for that channel, as an :obj:`int`. Refer to the
        documentation of the board to get the correspondence between range
        indexes and Volts. If not given, all input channels will be set to the
        range `0`.
      gain: A :obj:`list` containing for each input channel the gain to apply
        to the measured voltage, as a :obj:`float`. The returned voltage is
        calculated as follows :
        ::

          returned_voltage = gain * measured_voltage + offset

        If not given, no gain is applied to the measured values.
      offset: A :obj:`list` containing for each input channel the offset to
        apply to the measured voltage, as a :obj:`float`. The returned voltage
        is calculated as follows :
        ::

          returned_voltage = gain * measured_voltage + offset

        If not given, no offset is applied to the measured values.
      make_zero: A :obj:`list` containing for each input channel a :obj:`bool`
        indicating whether the channel should be zeroed or not. If so, data
        will be acquired on this channel before the test starts, and a
        compensation value will be deduced so that the offset of this channel
        is `0`. **It will only take effect if the ``make_zero_delay`` argument
        of the :ref:`IOBlock` controlling the Comedi is set** ! If not given,
        the channels are by default not zeroed.
      out_subdevice: The id of the subdevice to use for output channels, as an
        :obj:`int`.
      out_channels: A :obj:`list` containing the indexes of the channels to use
        as outputs, given as :obj:`int`.
      out_range_num: A :obj:`list` containing for each output channel the index
        of the range to set for that channel, as an :obj:`int`. Refer to the
        documentation of the board to get the correspondence between range
        indexes and Volts. If not given, all output channels will be set to the
        range `0`.
      out_gain: A :obj:`list` containing for each output channel the gain to
        apply to the command voltage, as a :obj:`float`. The set voltage is
        calculated as follows :
        ::

          set_voltage = out_gain * command_voltage + out_offset

        If not given, no gain is applied to the command values.
      out_offset: A :obj:`list` containing for each output channel the offset
        to apply to the command voltage, as a :obj:`float`. The set voltage is
        calculated as follows :
        ::

          set_voltage = out_gain * command_voltage + out_offset

        If not given, no offset is applied to the command values.

    Note:
      All the :obj:`list` given as arguments for the input channels should have
      the same length, and same for the output channels. If that's not the
      case, all the given lists are treated as if they had the same length
      as the shortest given list.
    """

    super().__init__()

    self._device_name = device.encode()
    self._subdevice = subdevice
    self._out_subdevice = out_subdevice

    # Setting the defaults for arguments that are not given
    if channels is None:
      channels = list()
    if out_channels is None:
      channels = list()

    if range_num is None:
      range_num = [0 for _ in channels]
    if gain is None:
      gain = [1 for _ in channels]
    if offset is None:
      offset = [0 for _ in channels]
    if make_zero is None:
      make_zero = [False for _ in channels]

    if out_range_num is None:
      out_range_num = [0 for _ in out_channels]
    if out_gain is None:
      out_gain = [1 for _ in out_channels]
    if out_offset is None:
      out_offset = [0 for _ in out_channels]

    # Creating the channel objects
    self._channels = [_Channel(num=chan, range_num=r_num, gain=g,
                               offset=off, make_zero=make_z)
                      for chan, r_num, g, off, make_z in
                      zip(channels, range_num, gain, offset, make_zero)]

    self._out_channels = [_Channel(num=chan, range_num=r_num, gain=g,
                                   offset=off) for chan, r_num, g, off in
                          zip(out_channels, out_range_num,
                              out_gain, out_offset)]

    self._device = None

  def open(self) -> None:
    """Opening the Comedi board and setting up the input and output
    channels."""

    # Opening the Comedi board
    self._device = comedi.comedi_open(self._device_name)

    # Setting up the input channels
    for chan in self._channels:
      chan.max_data = comedi.comedi_get_maxdata(self._device, self._subdevice,
                                                chan.num)
      chan.range_ds = comedi.comedi_get_range(self._device, self._subdevice,
                                              chan.num, chan.range_num)

    # Setting up the output channels
    for chan in self._out_channels:
      chan.max_data = comedi.comedi_get_maxdata(self._device,
                                                self._out_subdevice,
                                                chan.num)
      chan.range_ds = comedi.comedi_get_range(self._device,
                                              self._out_subdevice,
                                              chan.num, chan.range_num)

  def make_zero(self, delay: float) -> None:
    """Overriding of the method of the parent class, because the user can
    choose which channels should be zeroed or not.

    It simply performs the regular zeroing, and resets the compensation to
    zero for the channels that shouldn't be zeroed.
    """

    # No need to acquire data if no channel should be zeroed
    if any(chan.make_zero for chan in self._channels):

      # Acquiring the data
      super().make_zero(delay)

      # Proceed only if the acquisition went fine
      if self._compensations:

        # Resetting the compensation for channels that shouldn't be zeroed
        self._compensations = [comp if chan.make_zero else 0 for comp, chan
                               in zip(self._compensations, self._channels)]

  def set_cmd(self, *cmd: float) -> None:
    """Sets the command value on the output channels.

    There should be as many commands as there are output channels. In case
    there would be fewer commands or channels, the extra commands/channels
    wouldn't be considered/set.
    """

    for val, chan in zip(cmd, self._out_channels):

      # Adjusting with the provided gain and offset
      val = val * chan.gain + chan.offset

      # Converting voltage to numeric
      out_a = comedi.comedi_from_phys(val, chan.range_ds, chan.max_data)

      # Sending the command
      comedi.comedi_data_write(self._device, self._out_subdevice, chan.num,
                               chan.range_num, comedi.AREF_GROUND, out_a)

  def get_data(self) -> List[float]:
    """Simply reads and returns the value of each channel, adjusted with the
    given gain and offset."""

    data = [time()]

    for chan in self._channels:

      # Reading the numeric values
      data_read = comedi.comedi_data_read(self._device, self._subdevice,
                                          chan.num, chan.range_num,
                                          comedi.AREF_GROUND)

      # Converting numeric to a voltage
      val = comedi.comedi_to_phys(data_read, chan.range_ds, chan.max_data)

      # Adjusting with the provided gain and offset
      data.append(val * chan.gain + chan.offset)
    return data

  def close(self) -> None:
    """Simply closes the Comedi board and warns the user in case of failure."""

    if comedi.comedi_close(self._device):
      print('[Comedi] Device close failed !')



# coding: utf-8

from multiprocessing import Process, Pipe, RLock
from multiprocessing.connection import Connection
import multiprocessing.synchronize
from _io import FileIO
from tempfile import TemporaryFile
from typing import List, Dict, Any
from ..._global import OptionalModule
try:
  from usb.core import find, Device, USBTimeoutError
  from usb import util
except (FileNotFoundError, ModuleNotFoundError):
  find = OptionalModule('pyusb')
  Device = OptionalModule('pyusb')
  USBTimeoutError = OptionalModule('pyusb')
  util = OptionalModule('pyusb')


class Server_process(Process):
  """Process actually communicating with the FT232H device.

  It receives the commands, sends them to the device, and returns the answer.
  This architecture is necessary as :mod:`pyusb` doesn't support
  multiprocessing.
  """

  def __init__(self,
               new_block_recv: Connection,
               current_file: FileIO,
               command_file: FileIO,
               answer_file: FileIO,
               lock_pool: List[multiprocessing.synchronize.RLock],
               current_lock: multiprocessing.synchronize.RLock,
               dev_dict: Dict[str, Any]):
    """Simply passes the args as instance attributes.

    Args:
      new_block_recv: A pipe connection through which new blocks send
        information.
      current_file: A temporary file in which the index of the block currently
        communicating with the server is written.
      command_file: A temporary file containing the command the server has to
        send to the ft232h.
      answer_file: A temporary file containing the answer from the device after
        a command was sent.
      lock_pool: A :obj:`list` of RLocks, with each one affected to a different
        block. They indicate to the server that a command is ready, or to the
        block that an answer is ready.
      current_lock: A unique RLock that determines which block has control over
        the server. The different blocks all try to acquire this lock.
      dev_dict: A :obj:`dict` whose keys are the serial numbers of the
        connected ft232h and values are the associated :mod:`pyusb` Device
        objects.
    """

    super().__init__()

    self.new_block_recv = new_block_recv
    self.current_file = current_file
    self.command_file = command_file
    self.answer_file = answer_file
    self.lock_pool = lock_pool
    self.current_lock = current_lock
    self.dev_dict = dev_dict

  def _send_command(self,
                    command: bytes,
                    device,
                    serial_nr: str,
                    current_block: int) -> bytes:
    """Sends commands to the USB devices and returns the answer.

    Args:
      command: The command to send to the device. The bytes are arranged in a
        specific way for each type of command.
      device: The :mod:`pyusb` Device to which the commands are sent.
      serial_nr: The serial number of the :mod:`pyusb` device.
      current_block: The index of the block currently controlling the server.

    Returns:
      The number of the command, followed by the answer from the USB device if
      any.
    """

    command = command.split(b',')

    # Control transfer out
    if command[0] == b'00':
      return b'00,' + str(device.ctrl_transfer(int(command[1]),
                                               int(command[2]),
                                               int(command[3]),
                                               int(command[4]),
                                               command[5],
                                               int(command[6]))).encode()
    # Control transfer in
    elif command[0] == b'01':
      return b'01,' + bytes(device.ctrl_transfer(*[int(arg) for arg in
                                                   command[1:]]))
    # Write operation
    elif command[0] == b'02':
      return b'02,' + str(device.write(int(command[1]), command[2],
                                       int(command[3]))).encode()
    # Read operation
    elif command[0] == b'03':
        return b'03,' + bytes(device.read(*[int(arg) for arg in command[1:]]))
    # Checks whether the kernel driver is active
    # It doesn't actually interact with the device
    elif command[0] == b'04':
      return b'04,' + (b'1' if device.is_kernel_driver_active(int(command[1]))
                       else b'0')
    # Detaches the kernel driver
    # It doesn't actually interact with the device
    elif command[0] == b'05':
      device.detach_kernel_driver(int(command[1]))
      return b'05,'
    # Sets the device configuration
    elif command[0] == b'06':
      device.set_configuration()
      return b'06,'
    # Custom command getting information from the current configuration
    elif command[0] == b'07':
      info = self._return_config_info(device)
      return b'07,' + str(info[0]).encode() + b',' + str(info[1]).encode() + \
             b',' + str(info[2]).encode() + b',' + str(info[3]).encode()
    # When a block is leaving, if it's the last one associated with a given
    # ft232h then it should release the internal resources
    # It doesn't actually interact with the device
    elif command[0] == b'08':
      self.dev_count[serial_nr] -= 1
      return b'08,' + (b'1' if not self.dev_count[serial_nr] else b'0')
    # Checks whether the internal resources have been released or not
    # It doesn't actually interact with the device
    elif command[0] == b'09':
      return b'09,' + (b'1' if device._ctx.handle else b'0')
    # Releases the USB interface
    # It doesn't actually interact with the device
    elif command[0] == b'10':
      util.release_interface(device, int(command[1]))
      return b'10,'
    # Detaches the kernel driver
    # It doesn't actually interact with the device
    elif command[0] == b'11':
      device.attach_kernel_driver(int(command[1]))
      return b'11,'
    # Releases all the resources used by :mod:`pyusb` for a given device
    # It doesn't actually interact with the device
    elif command[0] == b'12':
      util.dispose_resources(device)
      return b'12,'
    # Registers a block as gone
    # It doesn't actually interact with the device
    elif command[0] == b'13':
      self.left[current_block] = True
      return b'13,'

  def run(self) -> None:
    """Main loop of the server.

    It runs an infinite loop for receiving the commands and sending back the
    answers to the blocks.
    """

    block_count = 0  # The count of blocks registered with the server
    num_to_dev = {}  # Associates each block index to a ft232h device
    num_to_ser = {}  # Associates each block index to a serial number
    self.left = {}  # For each block index, tells whether the block has left
    # Counts the number of blocks controlling a given device
    self.dev_count = {serial_nr: 0 for serial_nr in self.dev_dict}
    while True:
      try:
        # If all the blocks have left, stop the server
        if self.left and all(val for val in self.left.values()):
          break

        # A new block wants to register
        if self.new_block_recv.poll():
          try:
            # It sends a Connection object
            temp_pipe = self.new_block_recv.recv()
            if temp_pipe.poll(timeout=1):
              # The serial number of the ft232h the block wants to control is
              # sent
              serial_nr = temp_pipe.recv()
              # The different dicts are updated accordingly
              num_to_dev[block_count] = self.dev_dict[serial_nr]
              num_to_ser[block_count] = serial_nr
              self.dev_count[serial_nr] += 1
              self.left[block_count] = False
            else:
              raise TimeoutError("A block took too long to send the serial nr")
            # Sending back the index the block was assigned
            temp_pipe.send(block_count)
            block_count += 1
          except KeyboardInterrupt:
            # The program was interrupted during __init__, no choice but to
            # stop abruptly
            break

        # A block has acquired the lock and wrote its index in the file
        if self.current_file.tell() > 0:
          self.current_file.seek(0)
          # Reads the index of the block currently in control
          current_block = int(self.current_file.read())
          self.current_file.seek(0)
          self.current_file.truncate(0)

          # The block has finished writing the command
          if self.lock_pool[current_block].acquire(timeout=1):
            try:
              # Reads the command
              self.command_file.seek(0)
              command = self.command_file.read()
              self.command_file.seek(0)
              self.command_file.truncate(0)

              try:
                # Sends the command to the device and returns the answer
                answer = self._send_command(command,
                                            num_to_dev[current_block],
                                            num_to_ser[current_block],
                                            current_block)
              except (USBTimeoutError, TimeoutError):
                # Double-checking the timeout error
                answer = self._send_command(command,
                                            num_to_dev[current_block],
                                            num_to_ser[current_block],
                                            current_block)

              # Writing the answer in the file
              self.answer_file.seek(0)
              self.answer_file.truncate(0)
              try:
                self.answer_file.write(answer)
              except TypeError:
                # Sometimes for an unknown reason the answer is None
                raise KeyboardInterrupt

            except KeyboardInterrupt:
              # If the command wasn't sent it's no big deal, the block will
              # send it again anyway
              pass
            finally:
              # Releasing the lock so that the block can go on
              self.lock_pool[current_block].release()

      except KeyboardInterrupt:
        # The server should never raise exceptions, as it must keep running to
        # allow the blocks to finish properly
        continue

  @staticmethod
  def _return_config_info(device) -> tuple:
    """Returns some configuration information from a USB object.

    Args:
      device: A :obj:`usb.core.Device`

    Returns:
      The index, in endpoint, out endpoint and maximum packet size of a USB
      device
    """

    interface = device.get_active_configuration()[(0, 0)]
    index = interface.bInterfaceNumber + 1
    in_ep, out_ep = sorted([ep.bEndpointAddress for ep in interface])[:2]
    max_packet_size = interface[0].wMaxPacketSize
    return index, in_ep, out_ep, max_packet_size


class Usb_server:
  """Class for starting a server controlling communication with the
  :ref:`FT232H` devices.

  The :ref:`In / Out` objects wishing to communicate through an :ref:`FT232H`
  inherit from this class.

  Note:
    There is a limitation to 10 blocks accessing ft232h devices from the same
    machine in Crappy. This limit can be increased at will, but it is necessary
    to change the code of this class and build Crappy yourself.
  """

  def __init__(self, serial_nr: str, backend: str) -> None:
    """Simply receives the attributes from the :ref:`In / Out` object.

    Args:
      serial_nr (:obj:`int`): The serial number of the :ref:`FT232H` to use.
      backend (:obj:`str`): The server won't be started if the chosen backend
        is not ``'ft232h'``.
    """

    self._serial_nr = serial_nr
    self._backend = backend

  def start_server(self) -> tuple:
    """Starts the server for communicating with the :ref:`FT232H` devices.

    If the server is already started, doesn't start it twice. Then initializes
    the connection with the server and receives a block number.

    Returns:
      The different :mod:`multiprocessing` objects needed as arguments by the
      :ref:`FT232H` in order to run properly.
    """

    if self._backend == 'ft232h':
      # The server should only be started once
      if not hasattr(Usb_server, 'server'):
        # Finding all connected FT232H
        devices: List[Device] = list(find(find_all=True,
                                          idVendor=0x0403,
                                          idProduct=0x6014))
        if not devices:
          raise IOError("No FT232H connected")

        dev_dict = {}

        # Collecting all the serial numbers of the connected FT232H
        if len(devices) == 0:
          raise IOError("No FT232H connected")
        # If only one FT232H is connected then it is acceptable not to give a
        # serial number
        elif len(devices) == 1:
          if self._serial_nr == '':
            dev_dict[''] = devices[0]
          else:
            dev_dict[devices[0].serial_number] = devices[0]
        else:
          for device in devices:
            try:
              dev_dict[device.serial_number] = device
            except ValueError:
              if len(devices) > 1:
                raise ValueError('Please set a serial number for each FT232H '
                                 'using the corresponding crappy tool')
              else:
                dev_dict[''] = device

        # This pipe is used by each new block to send a Connection object
        # Through this connection the server receives the information of the
        # new block
        Usb_server.new_block_send, new_block_recv = Pipe()
        # In this file will be written the number of the block currently owning
        # the lock
        Usb_server.current_file = TemporaryFile(buffering=0)
        # In this file will be written the command from the block to the server
        Usb_server.command_file = TemporaryFile(buffering=0)
        # In this file will be written the answer from the server to the block
        Usb_server.answer_file = TemporaryFile(buffering=0)
        # Each new block is assigned one of these locks
        # It is used by the block to indicate the server that the command is
        # ready, and by the server to indicate the block that the answer is
        # ready
        Usb_server.lock_pool = [RLock() for _ in range(10)]
        # The block owning this lock is the only one that can communicate with
        # the server
        Usb_server.current_lock = RLock()

        # Starting the server process
        Usb_server.server = Server_process(
          new_block_recv, Usb_server.current_file,
          Usb_server.command_file,
          Usb_server.answer_file, Usb_server.lock_pool,
          Usb_server.current_lock, dev_dict)
        Usb_server.server.start()

      # Sending the server the serial number of the ft232h to use
      temp_pipe_block, temp_pipe_server = Pipe()
      Usb_server.new_block_send.send(temp_pipe_server)
      temp_pipe_block.send(self._serial_nr)
      if temp_pipe_block.poll(timeout=1):
        # Receiving the block number from the server
        self.block_number = temp_pipe_block.recv()
        self.block_lock = Usb_server.lock_pool[self.block_number]
      else:
        raise TimeoutError('The USB server took too long to reply')

      # Transmitting the necessary information to the InOut or Actuator object
      return Usb_server.current_file, self.block_number, \
          Usb_server.command_file, Usb_server.answer_file, \
          self.block_lock, Usb_server.current_lock

    return None, None, None, None, None, None

  def __del__(self) -> None:
    """Stops the server upon deletion of the :ref:`In / Out` object."""

    if hasattr(Usb_server, 'server') and Usb_server.server.is_alive():
      Usb_server.server.kill()



# coding: utf-8

"""
Example illustrating the use of Spectrum DAQ cards and high-freq saving.

The data is saved using the hdf format.

Required hardware:
  - Spectrum acquisition board
"""

import crappy

if __name__ == "__main__":
  channels = list(range(16))  # Every channel (from 0 to 15)
  ranges = [10000] * len(channels)  # -10/+10V (in mV)
  # This will NOT apply the gain to the stream, only save a key in the h5
  gains = [1] * len(channels)
  save_file = "./out.h5"

  chan_names = ['ch' + str(i) for i in channels]

  spectrum = crappy.blocks.IOBlock('spectrum', ranges=ranges,
                                   channels=channels,
                                   streamer=True,
                                   labels=['t(s)', 'stream'])

  graph = crappy.blocks.Grapher(*[('t(s)', i) for i in chan_names])

  if save_file:
    hrec = crappy.blocks.Hdf_recorder("./out.h5",
                                      metadata={
                                        'channels': channels,
                                        'ranges': ranges,
                                        'freq': 100000,
                                        'factor': [r * g / 32000000 for r, g in
                                                   zip(ranges, gains)]})
    crappy.link(spectrum, hrec)

  # The Demux modifier unpacks the stream data so normal blocks can process it
  crappy.link(spectrum, graph,
              modifier=crappy.modifier.Demux(chan_names, mean=False))
  # Note that the majority of the data is DROPPED with this modifier, it was
  # meant to plot the data in real time at a much lower rate
  # However, the hdf file contains the complete raw data

  crappy.start()



# coding: utf-8

from .camera import Camera
from .._global import OptionalModule
from time import time, sleep
from numpy import uint8, ndarray, uint16, copy, squeeze
from typing import Tuple, Optional, Union
from subprocess import Popen, PIPE, run
from platform import system
from re import findall, split, search

try:
  import cv2
except (ImportError, ModuleNotFoundError):
  pass
  cv2 = OptionalModule('opencv-python')

try:
  from gi import require_version
  require_version('Gst', '1.0')
  require_version('GstApp', '1.0')
  from gi.repository import Gst, GstApp
except (ImportError, ModuleNotFoundError, ValueError):
  Gst = GstApp = OptionalModule('gi', 'You need to install the python bindings'
                                      ' for GStreamer ! ')

# Todo: Error with a getter


class Camera_gstreamer(Camera):
  """A class for reading images from a video device using Gstreamer.

  It can read images from the default video source, or a video device can be
  specified. In this case, the user has access to a range of parameters for
  tuning the image. Alternatively, it is possible to give a custom GStreamer
  pipeline to the block. In this case no settings are available, and it is up
  to the user to ensure the validity of the pipeline.

  This class uses less resources and is compatible with more cameras than the
  :ref:`Camera OpenCV` camera, that relies on OpenCV. The installation of
  GStreamer of however less straightforward than the one of OpenCV.

  Note:
    For a better performance of this class in Linux, it is recommended to have
    `v4l-utils` installed.
  """

  def __init__(self) -> None:
    """Simply initializes the instance attributes."""

    super().__init__()
    Gst.init(None)

    self._last_frame_nr = 0
    self._frame_nr = 0
    self._img = None

    self._exposure_mode = None

  def open(self,
           device: Optional[Union[int, str]] = None,
           user_pipeline: Optional[str] = None,
           nb_channels: Optional[int] = None,
           img_depth: Optional[int] = None,
           **kwargs) -> None:
    """Opens the pipeline, sets the settings and starts the acquisition of
    images.

    A custom pipeline can be specified using the ``user_pipeline`` argument.
    Simply copy-paste the pipeline as it is in the terminal, and the class will
    take care of adjusting it to make it compatible with the Python binding
    of GStreamer.

    Note:
      For custom pipelines, it may be necessary to set a format directly in the
      pipeline (which isn't the case in the terminal). For instance, this line

      .. code-block:: shell-session

        gst-launch-1.0 autovideosrc ! videoconvert ! autovideosink

      May need to become

      .. code-block:: shell-session

        gst-launch-1.0 autovideosrc ! videoconvert ! video/x-raw,format=BGR ! \
autovideosink

    Note:
      Pipelines built using a pipe for redirecting the stream from another
      application are also accepted. For example, the following user pipeline
      is valid :

      .. code-block:: shell-session

        gphoto2 --stdout --capture-movie | gst-launch-1.0 fdsrc fd=0 ! \
videoconvert ! autovideosink

    Args:
      device: The video device to open, if the device opened by default isn't
        the right one. In Linux, should be a path like `/dev/video0`. In
        Windows and Mac, should be the index of the video device. This argument
        is ignored if a ``user_pipeline`` is given.
      user_pipeline (:obj:`str`, optional): A custom pipeline that can
        optionally be given. If given, the ``device`` argument is ignored. The
        pipeline should be given as it would be in a terminal.
      nb_channels (:obj:`int`, optional): The number of channels expected in
        the acquired images, in case a custom pipeline is given. Otherwise,
        this argument is ignored. For now, Crappy only manages 1- and 3-channel
        images.
      img_depth (:obj:`int`, optional): The bit depth of each channel of the
        acquired images, in case a custom pipeline is given. Otherwise, this
        argument is ignored. For now, Crappy only manages 8- and 16-bits deep
        images.
      **kwargs: Allows specifying values for the settings even before
        displaying the graphical interface.
    """

    # Checking the validity of the arguments
    if device is not None and system() == 'Linux' and not isinstance(device,
                                                                     str):
      raise ValueError("In Linux, device should be a string !")
    elif device is not None and system() in ['Darwin', 'Windows'] and not \
            isinstance(device, int):
      raise ValueError("In Windows and Mac, device should be an integer !")

    if img_depth is not None and img_depth not in [8, 16]:
      raise ValueError('The img_depth must be either 8 or 16 (bits)')

    if user_pipeline is not None and nb_channels is None:
      raise ValueError('nb_channels must be given if user_pipeline is !')
    if user_pipeline is not None and img_depth is None:
      raise ValueError('img_depth must be given if user_pipeline is !')

    self._process = None

    # Parsing the user pipeline if given
    if user_pipeline is not None:

      # Removing the call to gst-launch-1.0
      user_pipeline = user_pipeline.replace('gst-launch-1.0 ', '')
      # Splitting in case there's a pipe in the command
      user_pipeline = user_pipeline.split(' | ', 1)

      # There's no pipe in the pipeline
      if len(user_pipeline) == 1:
        user_pipeline = user_pipeline[0]

        # Setting the sink
        user_pipeline = user_pipeline.split(' ! ')
        user_pipeline[-1] = 'appsink name=sink'
        user_pipeline = ' ! '.join(user_pipeline)

      # There's a pipe in the pipeline
      else:
        user_application = user_pipeline[0]
        user_pipeline = user_pipeline[1]

        # Opening a subprocess handling the first half of the pipeline
        self._process = Popen(user_application.split(' '), stdout=PIPE)
        file_descriptor = self._process.stdout.fileno()

        # Setting the source and the sink, according to the file descriptor of
        # the subprocess
        user_pipeline = user_pipeline.split(' ! ')
        user_pipeline[0] = f'fdsrc fd={file_descriptor}'
        user_pipeline[-1] = 'appsink name=sink'
        user_pipeline = ' ! '.join(user_pipeline)

    # Setting the expected properties of the image
    self._device = device
    self._user_pipeline = user_pipeline
    self._nb_channels = nb_channels if nb_channels is not None else 3
    self._img_depth = img_depth if img_depth is not None else 8

    # Defining the settings in case no custom pipeline was given
    if user_pipeline is None:

      # First, determine if the exposure can be set (Linux only)
      if system() == "Linux":

        # Trying to run v4l2-ctl to get the available settings
        command = ['v4l2-ctl', '-l'] if device is None \
          else ['v4l2-ctl', '-d', device, '-l']
        try:
          check = run(command, capture_output=True, text=True)
        except FileNotFoundError:
          print("\n#######\n"
                "Warning ! The performance of the Camera_gstreamer "
                "class could be improved if v4l-utils was installed !"
                "\n#######\n")
          check = None
        check = check.stdout if check is not None else ''

        # Trying to find the exposure parameters in the returned string
        expo = search(r'exposure\s0x[\dA-Fa-f]+\s\(\w+\)\s+:\s'
                      r'min=(?P<min>\d+)\smax=(?P<max>\d+)', check)
        expo_auto = search(r'exposure_auto\s0x[\dA-Fa-f]+\s\(\w+\)\s+:\s'
                           r'min=(?P<min>\d+)\smax=(?P<max>\d+)', check)
        expo_abso = search(r'exposure_absolute\s0x[\dA-Fa-f]+\s\(\w+\)\s+:\s'
                           r'min=(?P<min>\d+)\smax=(?P<max>\d+)', check)

        # If there's an exposure parameter, getting its upper and lower limits
        if expo:
          expo_min, expo_max = map(int, expo.groups())
          self._exposure_mode = 'direct'

        # If there's an exposure parameter, getting its upper and lower limits
        elif expo_auto and expo_abso:
          expo_min, expo_max = map(int, expo_abso.groups())
          self._exposure_mode = 'auto'

        else:
          expo_min = None
          expo_max = None

        # Creating the parameter if applicable
        if expo_min is not None:
          self.add_scale_setting(name='exposure', lowest=expo_min,
                                 highest=expo_max, getter=self._get_exposure,
                                 setter=self._set_exposure)

      # Then, trying to get the available image encodings and formats
      if system() == 'Linux':

        self._formats = []

        # Trying to run v4l2-ctl to get the available formats
        command = ['v4l2-ctl', '--list-formats-ext'] if device is None \
            else ['v4l2-ctl', '-d', device, '--list-formats-ext']
        try:
          check = run(command, capture_output=True, text=True)
        except FileNotFoundError:
          check = None
        check = check.stdout if check is not None else ''

        # Splitting the returned string to isolate each encoding
        if findall(r'\[\d+]', check):
          check = split(r'\[\d+]', check)[1:]
        elif findall(r'Pixel\sFormat', check):
          check = split(r'Pixel\sFormat', check)[1:]
        else:
          check = []

        if check:
          for img_format in check:
            # For each encoding, finding its name
            name, *_ = search(r"'(\w+)'", img_format).groups()
            sizes = findall(r'\d+x\d+', img_format)

            # For each name, finding the available sizes
            for size in sizes:
              self._formats.append(f'{name} {size}')

        else:
          # If v4l-utils is not installed, proposing two encodings without
          # further detail
          self._formats = ['Default', 'MJPG']

      # For Windows and Mac proposing two encodings without further detail
      else:
        self._formats = ['Default', 'MJPG']

      # Finally, creating the parameter if applicable
      if self._formats:
        # The format integrates the size selection
        if ' ' in self._formats[0]:
          self.add_choice_setting(name='format',
                                  choices=tuple(self._formats),
                                  getter=self._get_format,
                                  setter=self._set_format)
        # The size is independent of the format
        else:
          self.add_choice_setting(name='format', choices=tuple(self._formats),
                                  setter=self._set_format)

      # These settings are always available no matter the platform
      self.add_choice_setting(name="channels", choices=('1', '3'), default='1')
      self.add_scale_setting(name='brightness', lowest=-1., highest=1.,
                             setter=self._set_brightness, default=0.)
      self.add_scale_setting(name='contrast', lowest=0., highest=2.,
                             setter=self._set_contrast, default=1.)
      self.add_scale_setting(name='hue', lowest=-1., highest=1.,
                             setter=self._set_hue, default=0.)
      self.add_scale_setting(name='saturation', lowest=0., highest=2.,
                             setter=self._set_saturation, default=1.)

    # Setting up GStreamer and the callback
    self._pipeline = Gst.parse_launch(self._get_pipeline())
    self._app_sink = self._pipeline.get_by_name('sink')
    self._app_sink.set_property("emit-signals", True)
    self._app_sink.connect("new-sample", self._on_new_sample)

    # Starting image acquisition
    self._pipeline.set_state(Gst.State.PLAYING)

    # Checking that images are read as expected
    t0 = time()
    while not self._frame_nr:
      if time() - t0 > 2:
        raise TimeoutError(
          "Waited too long for the first image ! There is probably an error "
          "in the pipeline. The format of the received images may be "
          "unexpected, in which case you can try specifying it.")
      sleep(0.01)

    # Setting the kwargs if any
    self.set_all(**kwargs)

  def get_image(self) -> Tuple[float, ndarray]:
    """Reads the last image acquired from the camera.

    Returns:
      The acquired image, along with a timestamp.
    """

    # Assuming an image rate greater than 0.5 FPS
    # Checking that we don't return the same image twice
    t0 = time()
    while self._last_frame_nr == self._frame_nr:
      if time() - t0 > 2:
        raise TimeoutError("Waited too long for the next image !")
      sleep(0.01)

    self._last_frame_nr = self._frame_nr

    return time(), copy(self._img)

  def close(self) -> None:
    """Simply stops the image acquisition."""

    self._pipeline.set_state(Gst.State.NULL)

    # Closes the subprocess started in case a user pipeline containing a pipe
    # was given
    if self._process is not None:
      self._process.terminate()

  def _restart_pipeline(self,
                        pipeline: str,
                        exposure: Optional[int] = None) -> None:
    """Stops the current pipeline, redefines it, and restarts it.

    Args:
      pipeline (:obj:`str`): The new pipeline to use.
      exposure (:obj:`int`, optional): The new exposure value to set.
    """

    # Stops the previous pipeline
    self._pipeline.set_state(Gst.State.NULL)

    # Redefines the pipeline and the callbacks
    self._pipeline = Gst.parse_launch(pipeline)
    self._app_sink = self._pipeline.get_by_name('sink')
    self._app_sink.set_property("emit-signals", True)
    self._app_sink.connect("new-sample", self._on_new_sample)

    # There's an extra step to set the exposure
    if system() == "Linux" and exposure is not None and \
            self._exposure_mode is not None:
      # Two different ways of setting the exposure
      if self._exposure_mode == 'direct':
        extra_args = (f"controls,"
                      f"""exposure={exposure if exposure is not None 
                                else self.exposure}""")
      else:
        extra_args = (f"controls,exposure_auto=1,"
                      f"""exposure_absolute={exposure if exposure is not None
                                         else self.exposure}""")

      # Setting the exposure is done via the 'extra-controls' property of the
      # 'v4l2src' source
      self._app_source = self._pipeline.get_by_name('source')
      structure, _ = Gst.Structure.from_string(extra_args)
      self._app_source.set_property('extra-controls', structure)

    # Restarts the pipeline
    self._pipeline.set_state(Gst.State.PLAYING)

  def _get_pipeline(self,
                    brightness: Optional[float] = None,
                    contrast: Optional[float] = None,
                    hue: Optional[float] = None,
                    saturation: Optional[float] = None,
                    img_format: Optional[int] = None) -> str:
    """Method that generates a pipeline, according to the settings.

    If a user-defined pipeline was given, it will always be returned.

    Args:
      brightness: The brightness value to set, as a float between -1 and 1.
      contrast: The contrast value to set, as a float between 0 and 2.
      hue: The hue value to set, as a float between -1 and 1.
      saturation: The saturation value to set, as a float between 0 and 2.
      img_format: The image format to set, as a string containing the name of
        the encoding and optionally both the width and height in pixels.

    Returns:
      A pipeline matching the current settings values.
    """

    # Returns the custom pipeline if any
    if self._user_pipeline is not None:
      return self._user_pipeline

    # Choosing the source according to the platform
    if system() == "Linux":
      source = 'v4l2src'
    elif system() == 'Windows':
      source = 'ksvideosrc'
    elif system() == 'Darwin':
      source = 'vfsvideosrc'
    else:
      source = 'autovideosrc'

    # The source argument is handled differently according to the platform
    if system() == 'Linux' and self._device is not None:
      device = f'device={self._device}'
    elif system() in ['Darwin', 'Windows'] and self._device is not None:
      device = f'device-index={self._device}'
    else:
      device = ''

    # Getting the format index
    img_format = img_format if img_format is not None else self.format

    try:
      format_name, img_size = img_format.split(' ')
    except ValueError:
      format_name, img_size = img_format, None

    # Adding a mjpeg decoder to the pipeline if needed
    img_format = '! jpegdec' if format_name == 'MJPG' else ''

    # Getting the width and height from the second half of the string
    if img_size is not None:
      width, height = map(int, img_size.split('x'))
    else:
      width, height = None, None

    # Including the dimensions in the pipeline
    img_size = f',width={width},height={height}' if width else ''

    # Finally, generate a single pipeline containing all the user settings
    return f"""{source} {device} name=source {img_format} ! videoconvert ! 
           video/x-raw,format=BGR{img_size} ! 
           videobalance 
           brightness={brightness if brightness is not None 
                       else self.brightness:.3f} 
           contrast={contrast if contrast is not None else self.contrast:.3f} 
           hue={hue if hue is not None else self.hue:.3f}
           saturation={saturation if saturation is not None 
                       else self.saturation:.3f} ! 
           appsink name=sink"""

  def _on_new_sample(self, app_sink):
    """Callback that reads every new frame and puts it into a buffer.

    Args:
      app_sink: The AppSink object containing the new frames.

    Returns:
      A GStreamer object indicating that the reading went fine.
    """

    sample = app_sink.pull_sample()
    caps = sample.get_caps()

    # Extracting the width and height info from the sample's caps
    height = caps.get_structure(0).get_value("height")
    width = caps.get_structure(0).get_value("width")

    # Getting the actual data
    buffer = sample.get_buffer()
    # Getting read access to the buffer data
    success, map_info = buffer.map(Gst.MapFlags.READ)
    if not success:
      raise RuntimeError("Could not map buffer data!")

    # Casting the data into a numpy array
    try:
      numpy_frame = ndarray(
        shape=(height, width, self._nb_channels),
        dtype=uint8 if self._img_depth == 8 else uint16,
        buffer=map_info.data)
    except TypeError:
      raise TypeError("Unexpected number of channels in the received image !\n"
                      "You can try adding something like ' ! videoconvert ! "
                      "video/x-raw,format=BGR ! ' before your sink to specify "
                      "the format.\n(here BGR would be for 3 channels)")

    # Converting to gray level if needed
    if self._user_pipeline is None and self.channels == '1':
      numpy_frame = cv2.cvtColor(numpy_frame, cv2.COLOR_BGR2GRAY)

    # Cleaning up the buffer mapping
    buffer.unmap(map_info)

    self._img = squeeze(numpy_frame)
    self._frame_nr += 1

    return Gst.FlowReturn.OK

  def _set_brightness(self, brightness) -> None:
    """Sets the image brightness."""

    self._restart_pipeline(self._get_pipeline(brightness=brightness))

  def _set_contrast(self, contrast) -> None:
    """Sets the image contrast."""

    self._restart_pipeline(self._get_pipeline(contrast=contrast))

  def _set_hue(self, hue) -> None:
    """Sets the image hue."""

    self._restart_pipeline(self._get_pipeline(hue=hue))

  def _set_saturation(self, saturation) -> None:
    """Sets the image saturation."""

    self._restart_pipeline(self._get_pipeline(saturation=saturation))

  def _set_exposure(self, exposure: int) -> None:
    """Sets the exposure using v4l2.

    Only works when the platform is Linux.
    """

    self._restart_pipeline(self._get_pipeline(), exposure=exposure)

  def _get_exposure(self) -> int:
    """Returns the current exposure value.

    Only works when the platform is Linux.
    """

    # Trying to run v4l2-ctl to get the exposure value
    if self._device is not None:
      command = ['v4l2-ctl', '-d', self._device, '-C',
                 'exposure_absolute' if self._exposure_mode == 'auto' else
                 'exposure']
    else:
      command = ['v4l2-ctl', '-C',
                 'exposure_absolute' if self._exposure_mode == 'auto' else
                 'exposure']
    try:
      expo = run(command, capture_output=True, text=True)
    except FileNotFoundError:
      expo = None

    # Extracting the exposure from the returned string
    expo = expo.stdout if expo is not None else ''
    expo = search(r'\d+', expo)
    if expo:
      return int(expo[0])
    else:
      raise IOError("Couldn't read exposure value from v4l2 !")

  def _set_format(self, img_format) -> None:
    """Sets the image encoding and dimensions."""

    self._restart_pipeline(self._get_pipeline(img_format=img_format))

  def _get_format(self) -> str:
    """Parses the v4l2-ctl -V command to get the current image format as an
    index."""

    # Sending the v4l2-ctl command
    if self._device is not None:
      command = ['v4l2-ctl', '-d', str(self._device), '-V']
    else:
      command = ['v4l2-ctl', '-V']
    check = run(command, capture_output=True, text=True).stdout

    # Parsing the answer
    format_ = width = height = ''
    if search(r"'(\w+)'", check) is not None:
      format_, *_ = search(r"'(\w+)'", check).groups()
    if search(r"(\d+)/(\d+)", check):
      width, height = search(r"(\d+)/(\d+)", check).groups()

    return f'{format_} {width}x{height}'



# coding: utf-8

from pkg_resources import resource_string, resource_filename
from numpy import frombuffer, uint8
from ._global import OptionalModule
from webbrowser import open

from . import actuator
from . import camera
from . import inout
from . import tool
from . import blocks
from . import links
from . import modifier
from .__version__ import __version__

# For compatibility (deprecated!)
condition = modifier


# Quick access to documentation
def doc():
  open('https://crappy.readthedocs.io/en/latest/')


# Useful aliases
link = links.link
Block = blocks.Block
stop = Block.stop_all
prepare = Block.prepare_all
launch = Block.launch_all
start = Block.start_all
renice = Block.renice_all
reset = Block.reset


# Data aliases
class resources:
  try:
    from cv2 import imdecode
    speckle = imdecode(frombuffer(resource_string('crappy',
                                                  'tool/data/speckle.png'),
                                  uint8), flags=0)

    ve_markers = imdecode(frombuffer(
      resource_string('crappy', 'tool/data/ve_markers.tif'), uint8), flags=0)

    pad = imdecode(frombuffer(resource_string('crappy', 'tool/data/pad.png'),
                              uint8), flags=0)
  except (ModuleNotFoundError, ImportError):
    speckle = OptionalModule('opencv-python')
    ve_markers = OptionalModule('opencv-python')
    pad = OptionalModule('opencv-python')

  paths = {'pad': resource_filename('crappy', 'tool/data/pad.png'),
           'speckle': resource_filename('crappy', 'tool/data/speckle.png'),
           've_markers': resource_filename('crappy',
                                           'tool/data/ve_markers.tif')}



# coding: utf-8

from time import time
from typing import Union, Dict, Optional
from itertools import cycle, islice

from .path import Path, condition_type


class Cyclic_ramp(Path):
  """The path cyclically alternates between two ramps with different slopes,
  based on two different conditions.

  It is equivalent to a succession of :ref:`ramp` paths.
  """

  def __init__(self,
               _last_time: float,
               _last_cmd: float,
               condition1: Union[str, condition_type],
               condition2: Union[str, condition_type],
               speed1: float,
               speed2: float,
               cycles: float = 1,
               init_value: Optional[float] = None) -> None:
    """Sets the args and initializes the parent class.

    The path always starts with ``speed1``, and then switches to ``speed2``.

    Args:
      _last_time: The last timestamp when a command was generated. For internal
        use only, do not overwrite.
      _last_cmd: The last sent command. For internal use only, do not
        overwrite.
      condition1: The condition for switching to ``speed2``. Refer to
        :ref:`generator path` for more info.
      condition2: The condition for switching to ``speed1``. Refer to
        :ref:`generator path` for more info.
      speed1: Slope of the first generated ramp, in `units/s`.
      speed2: Slope of the second generated ramp, in `units/s`.
      cycles: Number of cycles. Half cycles are accepted. If `0`, loops
        forever.
      init_value: If given, overwrites the last value of the signal as the
        starting point for the first ramp. In the specific case when this path
        is the first one in the list of dicts, this argument must be given !

    Note:
      ::

        [{'type': 'cyclic_ramp', 'speed1': 5, 'condition1': 'AIN0>2',
        'speed2': -2, 'condition2': 'AIN1<1', 'cycles': 5}]

      is equivalent to
      ::

        [{'type': 'ramp', 'speed': 5,'condition': 'AIN0>2'},
        {'type': 'ramp', 'value': -2, 'condition': 'AIN1<1'}] * 5
    """

    Path.__init__(self, _last_time, _last_cmd)

    if init_value is None and _last_cmd is None:
      raise ValueError('For the first path, an init_value must be given !')

    # Creates an interator object with a given length
    if cycles > 0:
      cycles = int(2 * cycles)
      self._speeds = islice(cycle((speed1, speed2)), cycles)
      self._conditions = islice(cycle((self.parse_condition(condition1),
                                       self.parse_condition(condition2))),
                                cycles)

    # Creates an endless iterator object
    else:
      self._speeds = cycle((speed1, speed2))
      self._conditions = cycle((self.parse_condition(condition1),
                                self.parse_condition(condition2)))

    # The current condition object and value
    self._condition = None
    self._speed = None

    # The last extreme command sent
    self._last_peak_cmd = _last_cmd if init_value is None else init_value

  def get_cmd(self, data: Dict[str, list]) -> float:
    """Returns the current value of the signal and raises :exc:`StopIteration`
    when the cycles are exhausted.

    Also manages the switch between the speeds and conditions 1 and 2.
    """

    # During the first loop, getting the first condition and speed
    if self._speed is None and self._condition is None:
      try:
        self._speed = self._speeds.__next__()
        self._condition = self._conditions.__next__()
      except StopIteration:
        raise

    # During other loops, getting the next condition and speed if the current
    # condition is met
    if self._condition(data):
      t = time()
      self._last_peak_cmd += self._speed * (t - self.t0)
      self.t0 = t
      try:
        self._speed = self._speeds.__next__()
        self._condition = self._conditions.__next__()
      except StopIteration:
        raise

    # Finally, returning the current value
    return self._last_peak_cmd + self._speed * (time() - self.t0)
